
from typing import Dict, Optional, Union
from . import DataManager
import pandas as pd
import datetime
import time
import io
import contextlib
from ..wrapper.mysql import RawDatabaseConnector, BasicDatabaseConnector, DerivedDatabaseConnector
from .fund_info_filter import filter_fund_info, fund_info_update
from ..struct import AssetPrice, FundScoreParam, TaaTunerParam, AssetTimeSpan
from .score import FundScoreManager, ScoreFunc
from .data_tables import FundDataTables
from ..view.raw_models import CSIndexComponent
from ..view.basic_models import *
from ..view.derived_models import *


@contextlib.contextmanager
def profiled(file_name=None):
    import cProfile
    import pstats

    pr = cProfile.Profile()
    pr.enable()
    yield
    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats('cumulative')
    ps.print_stats()
    # uncomment this to see who's calling what
    # ps.print_callers()
    if file_name is None:
        print(s.getvalue())
    else:
        open(file_name, 'w').write(s.getvalue())


class FundDataManager(DataManager):

    def __init__(self, start_time=None, end_time=None, fund_score_param: FundScoreParam = None, score_manager: FundScoreManager = None):
        DataManager.__init__(self,
            start_time or datetime.datetime(2005, 1, 1),
            end_time or datetime.datetime.now()
        )
        self._fund_score_param = fund_score_param or FundScoreParam(tag_type=1, score_method=1, is_full=1)
        self.dts = FundDataTables()
        self._is_inited = False
        self._score_manager = score_manager or FundScoreManager()
        self.set_score_param(self._fund_score_param)

    @property
    def inited(self):
        return self._is_inited

    def init(self, index_list=None, score_pre_calc=True, print_time=False, use_weekly_monthly_indicators=False, is_normal_type=False, is_mix_type=False):
        _tm_start = time.time()
        index_list = index_list or list(AssetTimeSpan.__dataclass_fields__.keys())

        def fetch_table(session, view):
            query = session.query(view)
            return pd.read_sql(query.statement, query.session.bind).drop(columns='_update_time')

        with RawDatabaseConnector().managed_session() as raw_session:
            _tm_raw_start = time.time()
            self.dts.cs_index_component = fetch_table(raw_session, CSIndexComponent)
            _tm_raw_cs_index_component = time.time()
            if print_time:
                print(f'\t[time][raw] fetch cs index component: {_tm_raw_cs_index_component - _tm_raw_start}')
 
        # info is necessary
        with BasicDatabaseConnector().managed_session() as quant_session:
            _tm_basic_start = time.time()
            self.dts.fund_info = fetch_table(quant_session, FundInfo)
            _tm_basic_fetch_info = time.time()
            self.dts.fund_info = fund_info_update(self.dts.fund_info)
            _filter_fund_info = filter_fund_info(self.dts.fund_info, index_list, is_normal_type, is_mix_type)
            _tm_basic_filter_info = time.time()
            self.dts.fund_list = set(_filter_fund_info.fund_id)
            self.dts.fund_index_map = {cur.fund_id: cur.index_id for cur in _filter_fund_info.itertuples()}
            self.end_date_dic = self.dts.fund_info[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']
            _tm_basic_prep_struct = time.time()
            self.dts.trading_days = fetch_table(quant_session, TradingDayList)
            self.dts.trading_days = self.dts.trading_days[self.dts.trading_days.datetime <= self.end_date]
            trim_trading_days = self.dts.trading_days[(self.dts.trading_days.datetime >= self.start_date) & (self.dts.trading_days.datetime <= self.end_date)].datetime
            _tm_basic_fetch_days = time.time()
            # index
            self.dts.index_info = fetch_table(quant_session, IndexInfo)
            _tm_basic_fetch_index_info = time.time()
            _index_query = quant_session.query(
                    IndexPrice.index_id,
                    IndexPrice.datetime,
                    IndexPrice.close
                ).filter(
                    IndexPrice.index_id.in_(index_list),
                    IndexPrice.datetime >= self.start_date,
                    IndexPrice.datetime <= self.end_date
                )
            self.dts.index_price = pd.read_sql(_index_query.statement, _index_query.session.bind).pivot_table(index='datetime', columns='index_id', values='close').reindex(trim_trading_days).fillna(method='ffill')
            _tm_basic_fetch_index_price = time.time()
            # fund nav
            _fund_nav_query = quant_session.query(
                    FundNav.fund_id,
                    FundNav.adjusted_net_value,
                    FundNav.unit_net_value,
                    FundNav.datetime
                    # FundNav.subscribe_status,
                    # FundNav.redeem_status,
                ).filter(
                    FundNav.fund_id.in_(self.dts.fund_list),
                    FundNav.datetime >= self.start_date,
                    FundNav.datetime <= self.end_date,
                )
            _nav_df = pd.read_sql(_fund_nav_query.statement, _fund_nav_query.session.bind)
            _tm_basic_fetch_fund_nav = time.time()

            # 这里我们制作一个总的pivot_table再取值，会比分别两次pivot_table快一些
            # 如果index-column pair有重复的值，pivot会报错而pivot_table不会，所以这里用pivot也许更好些
            # 但是pivot里边会调用set_index，导致datetime列的dtype从object变为datetime64[ns]
            _temp_fund_nav = _nav_df.pivot_table(index='datetime', columns='fund_id')
            self.dts.fund_nav = _temp_fund_nav['adjusted_net_value'].dropna(axis=1, how='all').fillna(method='ffill')
            self.dts.fund_unit_nav = _temp_fund_nav['unit_net_value'].dropna(axis=1, how='all').fillna(method='ffill')
            _tm_basic_pivot_fund_nav = time.time()

            # fund size and fund com hold
            # 机构持仓比例每六个月更新，提前取半年以上, 填充空白值用
            # fund_com_hold fund_size 和 fund_nav同结构，日期纵轴，基金代码横轴
            _fund_size_and_hold_rate_query = quant_session.query(
                    Fund_size_and_hold_rate.fund_id,
                    Fund_size_and_hold_rate.size,
                    Fund_size_and_hold_rate.institution_holds,
                    Fund_size_and_hold_rate.datetime
                ).filter(
                    Fund_size_and_hold_rate.fund_id.in_(self.dts.fund_list),
                    Fund_size_and_hold_rate.datetime >= self.start_date - datetime.timedelta(days=200),
                    Fund_size_and_hold_rate.datetime <= self.end_date,
                )
            _size_and_hold_rate = pd.read_sql(_fund_size_and_hold_rate_query.statement, _fund_size_and_hold_rate_query.session.bind)
            _dt_index = self.dts.trading_days[self.dts.trading_days.datetime >= (self.start_date - datetime.timedelta(days=200))].datetime
            _temp_size_and_hold_rate = _size_and_hold_rate.pivot(index='datetime', columns='fund_id').reindex(_dt_index).fillna(method='ffill')
            self.dts.fund_size = _temp_size_and_hold_rate['size']
            self.dts.fund_com_hold = _temp_size_and_hold_rate['institution_holds']
            _tm_basic_fetch_fund_size = time.time()

            if print_time:
                print(f'\t[time][basic] fetch fund info : {_tm_basic_fetch_info - _tm_basic_start}')
                print(f'\t[time][basic] filter fund info: {_tm_basic_filter_info - _tm_basic_fetch_info}')
                print(f'\t[time][basic] prep fund struct: {_tm_basic_prep_struct - _tm_basic_filter_info}')
                print(f'\t[time][basic] fetch trade days: {_tm_basic_fetch_days - _tm_basic_prep_struct}')
                print(f'\t[time][basic] fetch index info: {_tm_basic_fetch_index_info - _tm_basic_fetch_days}')
                print(f'\t[time][basic] fetch index price: {_tm_basic_fetch_index_price - _tm_basic_fetch_index_info}')
                print(f'\t[time][basic] fetch fund navs: {_tm_basic_fetch_fund_nav - _tm_basic_fetch_index_price}')
                print(f'\t[time][basic] pivot fund navs: {_tm_basic_pivot_fund_nav - _tm_basic_fetch_fund_nav}')
                print(f'\t[time][basic] fetch fund size: {_tm_basic_fetch_fund_size - _tm_basic_pivot_fund_nav}')
        _tm_basic = time.time()

        # get derived data first
        with DerivedDatabaseConnector().managed_session() as derived_session:
            _tm_deriv_start = time.time()
            _fund_indicator_query = derived_session.query(
                    FundIndicator.fund_id,
                    FundIndicator.datetime,
                    FundIndicator.alpha,
                    FundIndicator.beta,
                    FundIndicator.fee_rate,
                    FundIndicator.track_err,
                    FundIndicator.year_length,
                    FundIndicator.down_risk,
                    FundIndicator.info_ratio,
                    FundIndicator.ret_over_period,
                    FundIndicator.treynor,
                    FundIndicator.mdd,
                    FundIndicator.m_square,
                    FundIndicator.var,
                    FundIndicator.r_square,
                    FundIndicator.sharpe,
                    FundIndicator.annual_ret,
                    FundIndicator.annual_vol,
                    FundIndicator.time_ret,
                ).filter(
                    FundIndicator.fund_id.in_(self.dts.fund_list),
                    FundIndicator.datetime >= self.start_date,
                    FundIndicator.datetime <= self.end_date,
                )
            self.dts.fund_indicator = pd.read_sql(_fund_indicator_query.statement, _fund_indicator_query.session.bind)
            _tm_deriv_fetch_fund_indicator = time.time()
            self.dts.fund_indicator['index_id'] = self.dts.fund_indicator.fund_id.apply(lambda x: self.dts.fund_index_map[x])
            _tm_deriv_apply_fund_indicator = time.time()
            _fund_indicator_query = derived_session.query(
                FundIndicatorAnnual
            ).filter(
                FundIndicatorAnnual.fund_id.in_(self.dts.fund_list),
                FundIndicatorAnnual.datetime <= self.end_date,
            )
            self.dts.fund_indicator_annual = pd.read_sql(_fund_indicator_query.statement, _fund_indicator_query.session.bind)
            self.dts.fund_indicator_annual = self.dts.fund_indicator_annual.set_index(['fund_id','datetime'])
            _tm_deriv_fetch_fund_indicator_annual = time.time()
            # fetch index val data from derived table, and choose different pct value for different index_id
            _tm_deriv_fetch_index_pct = time.time()
            pct_index_list = list(TaaTunerParam.POOL.keys())
            _index_pct_query = derived_session.query(
                    IndexValuation
                ).filter(
                    IndexValuation.index_id.in_(pct_index_list),
                    IndexValuation.datetime >= self.start_date,
                    IndexValuation.datetime <= self.end_date,
                ).order_by(IndexValuation.datetime.asc())
            self._index_pct_df = pd.read_sql(_index_pct_query.statement, _index_pct_query.session.bind).set_index('index_id')
            # 按照trading date 作为index, fillna
            date_list = self.dts.trading_days.datetime.tolist()
            res = []
            for index_id in pct_index_list:
                df = self._index_pct_df[self._index_pct_df.index == index_id].set_index('datetime').reindex(date_list).reset_index().fillna(method='ffill')
                df.loc[:, 'index_id'] = index_id
                res.append(df)
            _index_pct_df = pd.concat(res, axis=0)
            self.dts.index_pct = _index_pct_df.pivot_table(index=['datetime', 'index_id'], values=['pe_pct', 'pb_pct', 'ps_pct'])
            self.dts.index_date_list = self.dts.index_pct.index.remove_unused_levels().levels[0]
            _tm_deriv_apply_index_pct = time.time()

            if use_weekly_monthly_indicators:
                # indicator weekly 只有周日有值 在每一个fund上fillna, merge week_ly into indicator
                _fund_indicator_weekly_query = derived_session.query(
                        FundIndicatorWeekly.fund_id,
                        FundIndicatorWeekly.datetime,
                        FundIndicatorWeekly.alpha_w,
                        FundIndicatorWeekly.beta_w,
                        FundIndicatorWeekly.track_err_w,
                    ).filter(
                        FundIndicatorWeekly.fund_id.in_(self.dts.fund_list),
                        FundIndicatorWeekly.datetime >= self.start_date,
                        FundIndicatorWeekly.datetime <= self.end_date,
                    )
                _fund_indicator_weekly = pd.read_sql(_fund_indicator_weekly_query.statement, _fund_indicator_weekly_query.session.bind)
                _fund_indicator_weekly = _fund_indicator_weekly.pivot_table(index='datetime', columns='fund_id')
                _dt = _fund_indicator_weekly.index.union(date_list).set_names('datetime')  # 交易日，和现有indicator index 周日做并集
                _fund_indicator_weekly = _fund_indicator_weekly.reindex(index=_dt).fillna(method='ffill').stack().swaplevel()
                _fund_indicator_weekly = _fund_indicator_weekly.groupby(level='fund_id').apply(lambda x: x.droplevel(level=0).loc[:self.end_date_dic[x.index.get_level_values(0).array[0]], :])
                _tm_process_indicator_weekly = time.time()

                # indicator monthly 只有每月最后一天有值 在每一个fund上fillna, merge monthly into indicator
                _fund_indicator_monthly = fetch_table(derived_session, FundIndicatorMonthly)
                _fund_indicator_monthly = _fund_indicator_monthly.pivot_table(index='datetime', columns='fund_id')
                _dt = _fund_indicator_monthly.index.union(date_list).set_names('datetime')  # 交易日，和现有indicator index做并集
                _fund_indicator_monthly = _fund_indicator_monthly.reindex(index=_dt).fillna(method='ffill').stack().swaplevel()
                _fund_indicator_monthly = _fund_indicator_monthly.groupby(level='fund_id').apply(lambda x: x.droplevel(level=0).loc[:self.end_date_dic[x.index.get_level_values(0).array[0]], :])
                _tm_process_indicator_monthly = time.time()
                # 本来join可以传一个list of df进去一起join，效率应该更高，但这里用了on参数指定fund_indicator的其中两列（而不是index），这样便只能一次传一个df进去（pandas不支持这种情况下传一个list）
                for df in [_fund_indicator_weekly, _fund_indicator_monthly]:
                    self.dts.fund_indicator = self.dts.fund_indicator.join(df, on=['fund_id', 'datetime'], how='left')
                _tm_process_indicator_join = time.time()
                self.dts.fund_indicator_annual.index.set_levels(self.dts.fund_indicator_annual.index.levels[1].date, level=1, inplace=True)
                the_dict = self.dts.fund_indicator_annual.to_dict(orient='index')
                _res = self.dts.fund_indicator[['fund_id', 'datetime']].apply(lambda x : self._make_annual_fac(the_dict, x), axis=1)
                _tm_process_fund_annual_apply= time.time()
                _annual_fac_df = pd.DataFrame(_res.tolist())
                self.dts.fund_indicator = pd.merge(self.dts.fund_indicator,_annual_fac_df,how='left',left_on=['fund_id','datetime'],right_on=['fund_id','datetime'])
                _tm_process_fund_annual_join = time.time()
            else:
                _tm_process_indicator_weekly = time.time()
                _tm_process_indicator_monthly = _tm_process_indicator_weekly
                _tm_process_indicator_join = _tm_process_indicator_weekly
                _tm_process_fund_annual_apply = _tm_process_indicator_weekly
                _tm_process_fund_annual_join = _tm_process_indicator_weekly

            if print_time:
                print(f'\t[time][deriv] fetch fund indicator: {_tm_deriv_fetch_fund_indicator - _tm_deriv_start}')
                print(f'\t[time][deriv] apply fund indicator: {_tm_deriv_apply_fund_indicator - _tm_deriv_fetch_fund_indicator}')
                print(f'\t[time][deriv] fetch fund indicator annual: {_tm_deriv_fetch_fund_indicator_annual - _tm_deriv_apply_fund_indicator}')
                print(f'\t[time][deriv] fetch index val: {_tm_deriv_fetch_index_pct - _tm_deriv_fetch_fund_indicator_annual}')
                print(f'\t[time][deriv] pivot raw index: {_tm_deriv_apply_index_pct - _tm_deriv_fetch_index_pct}')
                print(f'\t[time][deriv] process indicator weekly: {_tm_process_indicator_weekly - _tm_deriv_apply_index_pct}')
                print(f'\t[time][deriv] process indicator monthly: {_tm_process_indicator_monthly - _tm_process_indicator_weekly}')
                print(f'\t[time][deriv] process indicator join: {_tm_process_indicator_join - _tm_process_indicator_monthly}')
                print(f'\t[time][deriv] process indicator annually apply {_tm_process_fund_annual_apply - _tm_process_indicator_join}')
                print(f'\t[time][deriv] merge annual fac cost time {_tm_process_fund_annual_join - _tm_process_fund_annual_apply}')
        _tm_derived = time.time()
        self.dts.fix_data()
        _tm_fix = time.time()

        self._is_inited = True
        self._score_manager.set_dts(self.dts)
        if score_pre_calc:
            self._score_manager.pre_calculate(is_filter_c=True,use_weekly_monthly_indicators=use_weekly_monthly_indicators)
        _tm_finish = time.time()
        print(self.dts)
        if print_time:
            print(f'[time] basic: {_tm_basic - _tm_start}')
            print(f'[time] deriv: {_tm_derived - _tm_basic}')
            print(f'[time] fix_d: {_tm_fix - _tm_derived}')
            print(f'[time] score: {_tm_finish - _tm_fix}')
            print(f'[time] total: {_tm_finish - _tm_start}')

    def _make_annual_fac(self, the_dict, x):
        fund_id = x.fund_id
        dt = x.datetime
        dic = {'fund_id': fund_id, 'datetime': dt}
        last_year = datetime.date(dt.year - 1,12,31)
        last_two_year = datetime.date(dt.year - 2,12,31)

        try:
            dic['last_year_alpha'] = the_dict[(fund_id, last_year)]['alpha_annual']
            dic['last_year_beta'] = the_dict[(fund_id, last_year)]['beta_annual']
            dic['last_two_year_alpha'] = the_dict[(fund_id, last_two_year)]['alpha_annual']
            dic['last_two_year_beta'] = the_dict[(fund_id, last_two_year)]['beta_annual']
        except KeyError:
            pass
        return dic

    def _get_annual_alpha_beta(self, _fund_indicator_annual_fund_i, year):
        if year in _fund_indicator_annual_fund_i.index:
            return _fund_indicator_annual_fund_i.loc[year,'alpha_annual'], _fund_indicator_annual_fund_i.loc[year,'beta_annual']
        else:
            return None, None

    def set_score_param(self, score_param: FundScoreParam):
        self._score_manager.set_param(score_param)

    def get_index_pcts(self, dt):
        # jch: only pct within 7 days take effect
        INDEX_PCT_EFFECTIVE_DELAY_DAY_NUM = 7
        res = {}
        for index_id in self.dts.index_pct.columns:
            df = self.dts.index_pct[index_id]
            _filtered = df[(df.index <= dt) & (df.index >= dt - datetime.timedelta(days=INDEX_PCT_EFFECTIVE_DELAY_DAY_NUM))]
            if len(_filtered) > 0:
                res[index_id] = _filtered.iloc[-1]
        return res

    def get_index_pcts_df(self, dt):
        if dt not in self.dts.index_date_list:
            return pd.DataFrame()
        return self.dts.index_pct.loc[dt]

    def get_fund_score(self, dt, index_id, is_filter_c=True, score_func: Optional[ScoreFunc] = None) -> dict:
        return self._score_manager.get_fund_score(dt, index_id, is_filter_c, score_func)

    def get_fund_scores(self, dt, is_filter_c=True, fund_score_funcs: Optional[Dict[str, Union[ScoreFunc, str]]] = None) -> dict:
        return self._score_manager.get_fund_scores(dt, self.dts.index_list, is_filter_c, fund_score_funcs)

    def get_fund_nav(self, dt):
        df = self.dts.fund_nav
        return df.loc[dt].to_dict()

    def get_fund_unit_nav(self, dt):
        df = self.dts.fund_unit_nav
        return df.loc[dt].to_dict()

    def get_fund_purchase_fees(self):
        return self.dts.fund_info.set_index('fund_id').purchase_fee.to_dict()

    def get_fund_redeem_fees(self):
        return self.dts.fund_info.set_index('fund_id').redeem_fee.to_dict()

    def get_fund_info(self):
        return self.dts.fund_info

    def get_trading_days(self):
        return self.dts.trading_days.copy()

    def get_index_price(self, dt=None):
        if dt:
            return self.dts.index_price.loc[dt]
        else:
            return self.dts.index_price.copy()

    def get_index_price_data(self, dt):
        _index_price = self.get_index_price(dt)
        return AssetPrice(**_index_price.to_dict())

    def get_cs_index_component(self) -> pd.DataFrame:
        return self.dts.cs_index_component

    def get_fund_indicator_annual(self) -> pd.DataFrame:
        return self.dts.fund_indicator_annual

    def white_and_black_list_filter(self, score, score_raw, disproved_list):
        return self._score_manager._white_and_black_list_filter(score, score_raw, disproved_list)


def test():
    # m = FundDataManager('20190101', '20200101', score_manager=FundScoreManager())
    m = FundDataManager('20190101', '20200101')
    m.init()
    print(m.get_fund_info())
    print(m.get_trading_days())
    print(m.get_fund_score(datetime.date(2019, 1, 3), 'hs300'))


if __name__ == '__main__':
    with profiled(file_name='/tmp/test.txt'):
        m = FundDataManager('20190101', '20200101', score_manager=FundScoreManager())
        m.init()



import argparse
import sys

HELP_MSG = '''
surfing <command> [<args>]

The most commonly used commands are:
    init_raw                setup raw database
    init_basic              setup basic database
    init_derived            setup derived database
    init_view               setup view database
    download_raw            download raw data
    update_basic            update basic data base on raw data
    update_derived          update derived data based on basic and raw data
    update_view             update view data based on basic and derived data
    data_update             run download_raw, update_basic, update_derived and update_view in turn
    check_fund_list         check fund list to get new funds and delisted funds
'''


class SurfingEntrance(object):
    # TODO: rename to Surfing and move parser out of __init__
    # Make this class usable in other scripts

    def __init__(self):
        parser = argparse.ArgumentParser(description='Fund Command Tool', usage=HELP_MSG)
        parser.add_argument('command', help='Subcommand to run')
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print("Unrecognized command")
            parser.print_help()
            exit(1)
        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()

    def _get_yesterday(self):
        import datetime
        yesterday = datetime.datetime.today() - datetime.timedelta(days = 1)
        yesterday_str = datetime.datetime.strftime(yesterday, '%Y%m%d')
        print(f'yesterday: {yesterday_str}')
        return yesterday_str

    def datetime_test(self):
        import datetime
        print(datetime.datetime.today())

    def init_raw(self):
        from .data.wrapper.mysql import RawDatabaseConnector
        from .data.view.raw_models import Base

        print('Begin to create tables for database: raw')
        Base.metadata.create_all(bind=RawDatabaseConnector().get_engine())
        print('done')

    def init_basic(self):
        from .data.wrapper.mysql import BasicDatabaseConnector
        from .data.view.basic_models import Base

        print('Begin to create tables for database: basic')
        Base.metadata.create_all(bind=BasicDatabaseConnector().get_engine())
        print('done')

    def init_derived(self):
        from .data.wrapper.mysql import DerivedDatabaseConnector
        from .data.view.derived_models import Base

        print('Begin to create tables for database: derived')
        Base.metadata.create_all(bind=DerivedDatabaseConnector().get_engine())
        print('done')

    def init_view(self):
        from .data.wrapper.mysql import ViewDatabaseConnector
        from .data.view.view_models import Base

        print('Begin to create tables for database: view')
        Base.metadata.create_all(bind=ViewDatabaseConnector().get_engine())
        print('done')

    def download_raw(self, yesterday=None):
        if not yesterday:
            yesterday = self._get_yesterday()

        from .data.fund.raw.raw_data_downloader import RawDataDownloader
        from .util.config import SurfingConfigurator

        rq_license = SurfingConfigurator().get_license_settings('rq')
        raw_data_downloader = RawDataDownloader(rq_license)
        return raw_data_downloader.download(yesterday, yesterday), raw_data_downloader.get_updated_count()

    def update_basic(self, yesterday=None):
        if not yesterday:
            yesterday = self._get_yesterday()

        from .data.fund.basic.basic_data_processor import BasicDataProcessor

        basic_data_processor = BasicDataProcessor()
        return basic_data_processor.process_all(yesterday, yesterday), basic_data_processor.get_updated_count()

    def update_derived(self, yesterday=None):
        if not yesterday:
            yesterday = self._get_yesterday()

        from .data.fund.derived.derived_data_processor import DerivedDataProcessor
        derived_data_processor = DerivedDataProcessor()
        return derived_data_processor.process_all(yesterday, yesterday), derived_data_processor.get_updated_count()

    def update_view(self, yesterday=None):
        if not yesterday:
            yesterday = self._get_yesterday()

        from .data.fund.view.view_data_processor import ViewDataProcessor
        return ViewDataProcessor().process_all(yesterday, yesterday)

    def data_update(self):
        print('Start data update')

        yesterday = self._get_yesterday()

        failed_tasks = {}
        updated_count = {}

        print('Step 1. Start raw data download')
        raw_failed_tasks, raw_updated_count = self.download_raw(yesterday)
        if 'em_tradedates' in raw_failed_tasks:
            print('Step 1. No raw data to download')
            print(f'Done data update, {yesterday} is not trading day')
            return
        if len(raw_failed_tasks) > 0:
            failed_tasks['raw'] = raw_failed_tasks
        updated_count['raw'] = raw_updated_count
        print('Step 1. Done raw data download')

        print('Step 2. Start update basic')
        basic_failed_tasks, basic_updated_count = self.update_basic(yesterday)
        if len(basic_failed_tasks) > 0:
            failed_tasks['basic'] = basic_failed_tasks
        updated_count['basic'] = basic_updated_count
        print('Step 2. Done update basic')

        print('Step 3. Start update derived')
        derived_failed_tasks, derived_updated_count = self.update_derived(yesterday)
        if len(derived_failed_tasks) > 0:
            failed_tasks['derived'] = derived_failed_tasks
        updated_count['derived'] = derived_updated_count
        print('Step 3. Done update derived')

        print('Step 4. Start update view')
        view_failed_tasks = self.update_view(yesterday)
        if len(view_failed_tasks) > 0:
            failed_tasks['view'] = view_failed_tasks
        print('Step 4. Done update view')

        print('Done data update')
        print(f'Failed tasks:\n{failed_tasks}')
        print(f'Updated count:\n{updated_count}')

        from .util.wechat_bot import WechatBot
        wechat_bot = WechatBot()
        wechat_bot.send_data_update_status(yesterday, failed_tasks, updated_count)

        self.check_fund_list(yesterday, wechat_bot)

    def calc_fund_estimated_nav(self):
        import datetime
        from .data.fund.raw.raw_data_helper import RawDataHelper
        from .data.api.raw import RawDataApi
        from .util.config import SurfingConfigurator
        from .data.fund.raw.rq_raw_data_downloader import RqRawDataDownloader

        time_start = datetime.datetime.now()
        rq_license = SurfingConfigurator().get_license_settings('rq')
        rq_raw_data_downloader = RqRawDataDownloader(rq_license, RawDataHelper())
        stock_minute_df = rq_raw_data_downloader.stock_minute()
        # print(f'stock_minute_df\n {stock_minute_df}')
        if stock_minute_df is None or stock_minute_df.empty:
            # TODO: send alert if failed
            return False

        from .data.fund.basic.fund_estimated_nav_processor import FundEstimatedNavProcessor
        fund_estimated_nav_processor = FundEstimatedNavProcessor()
        fund_estimated_nav_processor.calc_fund_estimated_nav(stock_minute_df)
        time_end = datetime.datetime.now()
        print(f'Time for calc_fund_estimated_nav is {(time_end - time_start).total_seconds()} seconds')

    def check_fund_list(self, yesterday=None, wechat_bot=None):
        from .data.fund.raw.raw_data_helper import RawDataHelper
        from .data.api.raw import RawDataApi

        if yesterday is None:
            yesterday = self._get_yesterday()
        if wechat_bot is None:
            from .util.wechat_bot import WechatBot
            wechat_bot = WechatBot()
        new_funds_with_name = {}
        new_delisted_funds_with_name = {}
        new_funds, new_delisted_funds = RawDataHelper.get_new_and_delisted_fund_list(yesterday)
        if new_funds is not None and new_delisted_funds is not None:
            # 获取新增基金以及新摘牌基金列表
            agg_funds = new_funds.union(new_delisted_funds)
            if agg_funds:
                # 不是OF结尾的我们把它换成OF也去查一下
                agg_funds.update({one.split('.')[0] + '.OF' for one in agg_funds if not one.endswith('.OF')})

                api = RawDataApi()
                # 在已有的fund_info里找对应基金的信息
                wind_df = api.get_wind_fund_info(agg_funds)
                fund_name_got = {row.wind_id: row.desc_name for row in wind_df.itertuples(index=False)}
                # 在Choice的fund_info里找剩余还没找到的基金信息
                em_df = api.get_em_fund_info(agg_funds.difference(set(fund_name_got.keys())))
                fund_name_got.update({row.em_id: row.name for row in em_df.itertuples(index=False)})

                # 拆回到原来的两个列表里
                for fund_id in sorted(new_funds):
                    try:
                        new_funds_with_name[fund_id] = fund_name_got[fund_id]
                    except KeyError:
                        # 把结尾换成.OF再查一下
                        new_funds_with_name[fund_id] = fund_name_got.get(fund_id.split('.')[0] + '.OF', None)
                        if new_funds_with_name[fund_id] is not None:
                            new_funds_with_name[fund_id] += '(场内)'

                for fund_id in sorted(new_delisted_funds):
                    try:
                        new_delisted_funds_with_name[fund_id] = fund_name_got[fund_id]
                    except KeyError:
                        # 把结尾换成.OF再查一下
                        new_delisted_funds_with_name[fund_id] = fund_name_got.get(fund_id.split('.')[0] + '.OF', None)
                        if new_delisted_funds_with_name[fund_id] is not None:
                            new_delisted_funds_with_name[fund_id] += '(场内)'
        wechat_bot.send_new_fund_list(yesterday, new_funds_with_name, new_delisted_funds_with_name)


if __name__ == "__main__":
    entrance = SurfingEntrance()



# 数字货币期货与股票最大不同还是在仓位的计算上
# 不论是保证金计算还是市值计算都有区别,而这些都是体现在 PositionInfo 上的
# 因此,要想让回测引擎同时支持两种标的就需要在 PositionInfo 跟 PosItem 里写大量的 if
# 这在支持标的增多的时候会成为负担而且也不是一种简洁的方式
# 因此,我认为更合理的方式应是不同标的的 Pos 就用一个子类来实现,所以现在即使用了 if 实现之后也是需要重构的
# 而目前在可预计的范围内(两三个月?)这会测系统也不需要接入股票的,即使要接入别的标的也是数字货币的现货以及期权可能性大
# 所以现在采取的方法是直接改写 PositionInfo 使其只支持数字货币的期货
# 让程序先跑通再考虑怎么重构
import os
import yaml
from ..constant import OrderStatus, ExecRole, ExchangeID, Direction, OrderType, AssetType, PosiDirection
from collections import deque
import numpy as np
# 重写 structure 三部分,TransferRequest,虽然 order_manager和 backtest_engine 都出现了 transfer 相关的东西,但好像都不需要改
# 只需要改 position 里面的 transfer 就好了.好像也不需要重写 TransferRequest,只需要把里面的 ticker 理解方式改一下就好了
# 之前的 ticker 填的是具体要转的币,现在要转的是对应这个 ticker 的保证金


class PosItem:

    def __init__(self, **kwds):
        # 抛弃昌浩记录 posItem 的方式,他的方式是保证金也好合约也好都用 PosItem 记录
        # 但是保证金和合约很多attribute 都不一样,因此都记作 PosItem 不太合适,即使以后完善了也应该是分为不同种类的 position 的
        # 因此在这改变记法,初始化的时候就把保证金看做合约position的其中一个属性,因此 transfer 部分要进行改动
        # 现在采取的是开仓均价计算保证金的方式,因此保证金就有未实现与已实现两部分,而计算总保证金的时候是把两者都加一起计算
        self.exchange = kwds.get('exchange', ExchangeID.NOT_AVAILABLE)
        self.ticker = kwds.get('ticker')
        # 用来判断手续费率的,因为相同交易所不同到期日合约名称不一样,但其实是相同性质的合约,费率也都是一样
        self.fee_rate_type = kwds.get('fee_rate_type')
        # 记录这个合约是用什么作为保证金的,其实只支持倒数合约,因为在计算 margin 变化的时候只对这一种做支持
        # 所以在这两边期货合约保证金都是BTC
        self.margin_ticker = kwds.get('margin_ticker', PositionInfo.MARGIN)
        # 记录当前仓位的开仓均价
        self.avg_cost = kwds.get('avg_cost', 0)
        self.asset_type = kwds.get('asset_type', AssetType.NOT_AVAILABLE)
        self.posi_direction = kwds.get('posi_direction', PosiDirection.NET)
        # 假设可以开到无限杠杆吧,或者说维持保证金率是 0,也就是说只要保证金大于 0 就不爆仓,所以 margin_frozen 这项应该总为 0
        self.margin_frozen = kwds.get('margin_frozen', 0)
        # 初始化的时候相当于起始保证金,初始化时候必须得有,之后只有每次平仓的时候才会改变这个值,开仓的话是不会变的
        self.realized_margin = kwds.get('realized_margin', 0)
        if self.realized_margin is None:
            raise Exception("没钱炒什么币,充值再来")
        # 相当于浮盈浮亏,计算的是开仓了的那部分浮动盈亏,但为避免交易所插针或者 1token 数据有误,这里计算用的价格是近 100 个 tick 的 last 的算术平均值
        self.unrealized_margin = kwds.get('unrealized_margin', 0)
        self.cal_unreal_len = 100
        self.clear_prices = deque(maxlen=self.cal_unreal_len)
        self.mean_price = 0
        # 如果是合约的仓位,大于 0 代表多头,小于 0 代表空头
        self.position = kwds.get('position', 0)
        # 应该叫 margin_to_transfer,但能转账的只有保证金,所以这个就不改了
        self.to_transfer = kwds.get('to_transfer', 0)
        self.leverage = 0

    @staticmethod
    def ticker_to_exchange(ticker):
        return 'BITMEX' if len(ticker) > 10 else 'OKEX'

    def get_market_value(self):
        return self.realized_margin + self.unrealized_margin

    def transfer(self, volume):
        if self.realized_margin + self.unrealized_margin - self.margin_frozen - self.to_transfer >= volume:
            self.realized_margin -= volume
            self.to_transfer += volume
            return True
        return False

    def init_clear(self, tick):
        # clear_prices 为空时对其进行初始化
        self.clear_prices[tick.ticker] = deque(maxlen=self.cal_unreal_len)
        self.clear_prices[tick.ticker].append((tick.ask_price1 + tick.bid_price1) / 2)

    def append_clear(self, tick):
        # 后面把每个 tick 的价格都append 在后面,顺便把 unrealized 都更新一下
        self.clear_prices.append((tick.ask_price1 + tick.bid_price1) / 2)
        self.mean_price = np.mean(self.clear_prices)
        if self.position != 0:
            self.unrealized_margin = self.position * PositionInfo.CONTRACT_SIZE[self.ticker_to_exchange(tick.ticker)] * (1 / self.avg_cost - 1 / self.mean_price)
            self.leverage = abs(self.position * PositionInfo.CONTRACT_SIZE[self.ticker_to_exchange(tick.ticker)] / self.mean_price) / (self.realized_margin + self.unrealized_margin)

    def __str__(self):
        return '<PosItem ex={} ticker={} at={} d={} pos={} trans={} realized={} unrealized={} avg_cost={} leverage={}>'.format(
            ExchangeID.read(self.exchange), self.ticker, AssetType.read(self.asset_type),
            PosiDirection.read(self.posi_direction), self.position, self.to_transfer, self.realized_margin, self.unrealized_margin, self.avg_cost, self.leverage
        )


class PositionInfo:
    
    MARGIN = 'btc'
    CONTRACT_SIZE = {'BITMEX': 1,
                        'OKEX': 100}
    os.path.abspath('../config/fee_rate.yml')
    with open('../config/fee_rate.yml', 'r') as f:
        FEE_RATE = yaml.safe_load(f)
                
    def __init__(self, **kwds):
    
        
        _holdings = kwds.get('holdings', [])
        self.holdings = []
        self._tick_idx = {}
        for pi in _holdings:
            self.create_holding(pi)

    def create_holding(self, pos_item):
        self.holdings.append(pos_item)
        # 以 ticker 和 exchange 来标识这个仓位,ticker在这相当于合约的品种
        self._tick_idx[(pos_item.ticker, pos_item.exchange)] = len(self.holdings) - 1  # 以 ticker 和 exchange 来标识这个仓位

    def get_holding(self, ticker, exchange=ExchangeID.NOT_AVAILABLE):
        idx = self._tick_idx.get((ticker, exchange))
        return None if idx is None else self.holdings[idx]

    def init_transfer(self, transfer):
        from_pos = self.get_holding(ticker=transfer.from_ticker, exchange=transfer.from_exchange)
        return False if from_pos is None else from_pos.transfer(transfer.volume)

    def confirm_transfer(self, transfer):
        from_pos = self.get_holding(ticker=transfer.from_ticker, exchange=transfer.from_exchange)
        from_pos.to_transfer -= transfer.volume
        to_pos = self.get_holding(ticker=transfer.to_ticker, exchange=transfer.to_exchange)
        if to_pos is None:
            to_pos = PosItem(exchange=transfer.to_exchange, ticker=transfer.to_ticker, asset_type=transfer.asset_type)
            self.create_holding(to_pos)
        to_pos.realized_margin += transfer.volume

    @staticmethod
    def get_commission(trade, pos):
        # 根据 trade 计算手续费多少
        rate = PositionInfo.FEE_RATE[ExchangeID.read(trade.exchange)][AssetType.read(trade.asset_type)][pos.fee_rate_type][-1][ExecRole.read(trade.exec_role)]
        commission = rate * trade.trade_volume * PositionInfo.CONTRACT_SIZE[ExchangeID.read(trade.exchange)] / trade.trade_price
        return commission

    def process_trade(self, trade):
        # print(f'来到了 structure 的 process_trade,你的交易是{trade}')
        # 只处理数字货币的期货交易
        pos = self.get_holding(ticker=trade.ticker, exchange=trade.exchange)
        if pos is None:
            pos = PosItem(exchange=trade.exchange, ticker=trade.ticker, asset_type=trade.asset_type)
            self.create_holding(pos)
        # print(f'你要进行处理的仓位是{pos}')
        if pos.position < 0:
            # 原本空头,现在如果是 sell 则重新计算 avg_cost
            if trade.direction == Direction.SELL:
                new_position = pos.position - trade.trade_volume
                pos.avg_cost = new_position / (pos.position / pos.avg_cost - trade.trade_volume / trade.trade_price)
                pos.position = new_position
                # 手续费从已实现保证金改变
                pos.realized_margin -= self.get_commission(trade, pos)
            elif trade.direction == Direction.BUY:
                # 如果是平仓, 平均成本不变,但改变已实现盈亏
                if trade.trade_volume + pos.position > 0:
                    # 这时候仓位得反向了
                    pos.realized_margin -= self.get_commission(trade, pos)
                    pos.realized_margin += pos.position * PositionInfo.CONTRACT_SIZE[ExchangeID.read(trade.exchange)] * ((1 / pos.avg_cost) - (1 / trade.trade_price))
                    pos.position += trade.trade_volume
                    pos.avg_cost = trade.trade_price
                else:
                    pos.position += trade.trade_volume
                    pos.realized_margin += trade.trade_volume * PositionInfo.CONTRACT_SIZE[ExchangeID.read(trade.exchange)] * ((1 / trade.trade_price) - (1 / pos.avg_cost)) - self.get_commission(
                        trade, pos)
        elif pos.position > 0:
            if trade.direction == Direction.BUY:
                new_position = pos.position + trade.trade_volume
                pos.avg_cost = new_position / (pos.position / pos.avg_cost + trade.trade_volume / trade.trade_price)
                pos.position = new_position
                pos.realized_margin -= self.get_commission(trade, pos)
            elif trade.direction == Direction.SELL:
                if pos.position - trade.trade_volume < 0:
                    pos.realized_margin -= self.get_commission(trade, pos)
                    pos.realized_margin += pos.position * PositionInfo.CONTRACT_SIZE[ExchangeID.read(trade.exchange)] * ((1 / pos.avg_cost) - (1 / trade.trade_price))
                    pos.position -= trade.trade_volume
                    pos.avg_cost = trade.trade_price
                else:
                    pos.position -= trade.trade_volume
                    pos.realized_margin += trade.trade_volume * PositionInfo.CONTRACT_SIZE[ExchangeID.read(trade.exchange)] * ((1 / pos.avg_cost) - (1 / trade.trade_price)) - self.get_commission(
                        trade, pos)
        else:
            # 此时仓位是 0,第一次开仓状态
            pos.realized_margin -= self.get_commission(trade, pos)
            pos.avg_cost = trade.trade_price
            if trade.direction == Direction.BUY:
                pos.position = trade.trade_volume
            elif trade.direction == Direction.SELL:
                pos.position = -trade.trade_volume
            else:
                raise Exception(f'下单方向有问题{trade.direction}')
        # print(f'处理完后的仓位是-{pos}')

    def cal_unrealized(self, tick):
        for pi in self.holdings:
            if pi.ticker == tick.ticker:
                # 没办法,因为获取不了 tick 的 exchange 信息,所以无法直接用 get_holding
                pi.append_clear(tick)

    def get_total_market_value(self):
        total_market_value = 0
        for pi in self.holdings:
            total_market_value += pi.get_market_value()
        return total_market_value

    def __str__(self):
        msg = '<'
        for pi in self.holdings:
            msg += ' {}\n'.format(pi)
        msg += '>'
        return msg

class TransferRequest:
    # transfer 也需要改写一下,因为之前是直接转账一个 PosItem,但现在转的是 PosItem 中的margin(直接默认这两个 margin 相同吧)
    # 所以参数的 ticker 需要改为 from_ticker 和 to_ticker
    def __init__(self, **kwds):
        self.transfer_id = kwds.get('transfer_id')
        self.from_ticker = kwds.get('from_ticker')
        self.to_ticker = kwds.get('to_ticker')
        self.asset_type = kwds.get('asset_type', AssetType.NOT_AVAILABLE)
        self.from_exchange = kwds.get('from_exchange')
        self.to_exchange = kwds.get('to_exchange')
        self.volume = kwds.get('volume')
        self.source = kwds.get('source')
        self.err_msg = kwds.get('err_msg')

    def __str__(self):
        return f'<Transfer id={self.transfer_id} from_ticker={self.from_ticker} to_ticker={self.to_ticker} at={AssetType.read(self.asset_type)} from_exchange={self.from_exchange} to_exchange' \
               f'={self.to_exchange} volume={self.volume} source={self.source}>'




if __name__ == '__main__':
    print(PositionInfo.FEE_RATE)



from typing import List, Dict
import pandas as pd
import numpy as np
import datetime
import traceback
import time
from .derived_data_helper import DerivedDataHelper
from ...manager.manager_fund import FundDataManager
from ...view.derived_models import FundIndicatorAnnual


class FundIndicatorProcessorAnnual:

    _MONTHS_A_YEAR = 12

    def __init__(self, data_helper):
        self._data_helper: DerivedDataHelper = data_helper
        self._res: List[pd.DataFrame] = []
        self._res2: List[pd.DataFrame] = []
        self._result = None
        self._result2 = None

    def init(self, start_date: str, end_date: str, dm=None, resample_day='1M'):
        # DM作为基础数据
        if dm is None:
            self._dm = FundDataManager(start_time=start_date, end_time=end_date)
            self._dm.init(score_pre_calc=False)
        else:
            self._dm = dm
        self._resample_day = resample_day

        # index price, fund nav 日期对齐至交易日
        self._trading_days = self._dm.dts.trading_days[self._dm.dts.trading_days.datetime >= self._dm.start_date].datetime
        self._fund_nav = self._dm.dts.fund_nav.reindex(self._trading_days)
        index_price = self._dm.dts.index_price.reindex(self._trading_days)
        pd.testing.assert_index_equal(self._fund_nav.index, index_price.index)

        # fund id到index id的字典
        self._fund_id_to_index_id: Dict[str, str] = self._dm.dts.fund_info[['fund_id', 'index_id']].set_index('fund_id').to_dict()['index_id']

        # 超过基金终止日的基金净值赋空
        fund_to_enddate_dict = self._dm.dts.fund_info[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']
        for fund_id in self._fund_nav.columns:
            fund_end_date = fund_to_enddate_dict[fund_id]
            if self._dm.end_date > fund_end_date:
                self._fund_nav.loc[fund_end_date:, fund_id] = np.nan

        # resample and datetime convert
        self._fund_ret = self._df_resample(np.log(self._fund_nav / self._fund_nav.shift(1)))
        self._index_ret = self._df_resample(np.log(index_price / index_price.shift(1)))

        # fund ret上加一列index id
        self._fund_ret = self._fund_ret.stack().reset_index().rename(columns={0: 'ret'})
        self._fund_ret['index_id'] = self._fund_ret.fund_id.map(self._fund_id_to_index_id)
        # 过滤掉没有index_id的基金
        self._fund_ret = self._fund_ret[self._fund_ret.index_id.notna()]

    def _df_resample(self, df):
        # resample前需要转成datetime索引
        df = df.set_axis(pd.to_datetime(df.index), inplace=False)
        # min_count=1，使得全NA的列最后的和也为NA(否则为0)
        df = df.resample(self._resample_day).sum(min_count=1)
        return df

    def _lambda_1(self, x: pd.Series, fund: pd.DataFrame, benchmark: pd.Series):
        fund: pd.DataFrame = fund.loc[x, :]
        benchmark: pd.Series = benchmark.loc[x]
        # shape[0]与benchmark的基金才可以回归
        fund = fund.loc[:, fund.count() == benchmark.size]
        # 每一列是一只基金，一起回归计算alpha
        ploy_res = np.polyfit(x=benchmark, y=fund, deg=1)
        result = pd.DataFrame({'datetime': x.array[-1], 'alpha_annual': ploy_res[1, :] * self._MONTHS_A_YEAR, 'beta_annual': ploy_res[0, :]}, index=fund.columns)
        # 计算超额收益
        er = fund.sum() - benchmark.sum()
        self._res.append(result.assign(excess_return_annual=er.T))

        # # 另一种方法算的alpha和beta，可作验证用
        # beta = fund.apply(lambda x: benchmark.cov(x)) / benchmark.var()
        # # jensen alpha
        # ri = (fund.sum()) * 12 / len(fund)
        # rm = (benchmark.sum()) * 12 / len(benchmark)
        # jensen_alpha = ri - beta * rm
        # self._res2.append(pd.DataFrame({'date': fund.index[-1], 'alpha': jensen_alpha, 'beta': beta}))

    def _process_fund_indicator_item(self, fund_ret, index_id, index_ret):
        # resample的apply每次只能操作一个Series，所以这里我们只对索引做resample+apply，同时将df作为参数传进去
        # 另外因为在func参数中只有一个函数时只会返回Series，我们需要在里边自己存储计算结果
        fund_ret.index.to_series().resample('1Y').apply(self._lambda_1, fund=fund_ret, benchmark=index_ret)

    def calculate_history(self, index_id_list: List[str] = None):
        # 支持传进来一个index list，所以这里没有直接group by index id
        # 获得index list
        index_list = self._fund_ret.index_id.unique() if index_id_list is None else index_id_list
        for index_id in index_list:
            _tm_1 = time.time()
            # 获取对应该index的所有fund
            fund_ret = self._fund_ret[self._fund_ret.index_id == index_id].drop(columns='index_id')
            fund_ret = fund_ret.pivot_table(index='datetime', columns='fund_id', values='ret')
            # 获取benchmark
            index_ret = self._index_ret.loc[:, index_id]
            self._process_fund_indicator_item(fund_ret, index_id, index_ret)
            _tm_2 = time.time()
            print(f'index {index_id} fund number: {fund_ret.shape[1]} finished, cost time {_tm_2 - _tm_1}s')
        self._result = pd.concat(self._res).replace({-np.Inf:None,np.Inf:None})
        # self._result2 = pd.concat(self._res2)

    def process(self, year=3):
        failed_tasks = []
        try:
            now_date: datetime.date = datetime.datetime.now().date()
            # 起始时间为第一年的1月1日
            start_date = datetime.date(now_date.year - year + 1, 1, 1)
            start_date = start_date.strftime('%Y%m%d')
            if (now_date + datetime.timedelta(days=1)).month != now_date.month:
                # 如果是本月最后一天，直接使用这个日期
                end_date = now_date
            else:
                # 如果不是，取上月最后一天
                end_date = datetime.date(now_date.year, now_date.month, 1) - datetime.timedelta(days=1)
            end_date = end_date.strftime('%Y%m%d')
            print(start_date, end_date)
            self.init(start_date=start_date, end_date=end_date)
            self.calculate_history()
            self._data_helper._upload_derived(self._result.reset_index(), FundIndicatorAnnual.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_indicator_annual')
        return failed_tasks


if __name__ == '__main__':
    fipa = FundIndicatorProcessorAnnual(DerivedDataHelper())
    fipa.process()




from .structure import Tick, Order, Trade, OrderStatus, ExecRole, Direction, OrderType, PositionInfo
from .constant import AssetType
from queue import Queue

class MatchEngine:

    def __init__(self,run_engine=None,strategy_type=None):
        self.run_engine = run_engine
        self.ticks = {}
        self.pending_orders = {}
        self.id_orders = {}
        self._cur_trade_id = 1000000
        self.strategy_type = strategy_type
        self.stock_slippage = PositionInfo.FEE_RATE['STOCK']['SLIPPAGE']
 

    def init(self):
        if self.strategy_type != 'stock_daily':
            print(self.strategy_type)
            for ticker in self.run_engine._universe:
                self.pending_orders[ticker] = Queue()

    def process_order(self, order, on_tick=False):
        if order.status == OrderStatus.QUEUEING:
            # TODO: now we assume once traded, all traded, no partly traded
            trades = self.match_order(order, on_tick)
            if trades:
                for td in trades:
                    self.run_engine._push_trade(td)
                order.status = OrderStatus.ALL_TRADED
                self.run_engine._push_order_status(order)
                self.id_orders.pop(order.order_id, None)
            else:
                self.id_orders[order.order_id] = order
                self.pending_orders[order.ticker].put(order)

    def process_order_daily(self, order):
        trades = self.match_order_daily(order)
        if trades:
            for td in trades:# in case one order split into multiple trades
                self.run_engine._push_trade(td)

    def cancel_order(self, order_id):
        order = self.id_orders.get(order_id)
        if not order is None and order.status == OrderStatus.QUEUEING:
            order.status = OrderStatus.CANCELLED
            self.run_engine._push_order_status(order)
            self.id_orders.pop(order_id, None)
            return True
        else:
            return False

    def on_tick(self, tick):
        self.ticks[tick.ticker] = tick
        self.process_orders(tick.ticker, True)

    def on_day(self, day_data):
        self.day_data = day_data        

    def process_orders(self, ticker, on_tick=False):
        i = 0
        sz = self.pending_orders[ticker].qsize()
        # at most we loop over the whole pending orders once.
        while i < sz and not self.pending_orders[ticker].empty():
            od = self.pending_orders[ticker].get()
            self.process_order(od, on_tick)
            i += 1

    #############################
    # on_tick is a bool
    #   True: this matching is triggered by a tick data
    #   False: this matching is triggered by a order
    #############################
    def match_order(self, order, on_tick):
        tick = self.ticks[order.ticker]
        if tick is None:
            return None
        trade = Trade(order_id=order.order_id, exchange=order.exchange, ticker=order.ticker, asset_type=order.asset_type,
                      direction=order.direction, trade_volume=order.volume, source=order.source)
        if order.order_type is OrderType.MARKET:
            trade.trade_price = tick.ask_price1 if order.direction == Direction.BUY else tick.bid_price1
            trade.trade_id = self.new_trade_id()
            trade.exec_role = ExecRole.TAKER
            return [trade]
        else:
            if order.direction == Direction.BUY and order.price >= tick.ask_price1:
                trade.trade_price = order.price if on_tick else tick.ask_price1
                trade.exec_role = ExecRole.MAKER if on_tick else ExecRole.TAKER
                trade.trade_id = self.new_trade_id()
                return [trade]
            elif order.direction == Direction.SELL and order.price <= tick.bid_price1:
                trade.trade_price = order.price if on_tick else tick.bid_price1
                trade.exec_role = ExecRole.MAKER if on_tick else ExecRole.TAKER
                trade.trade_id = self.new_trade_id()
                return [trade]
        return None

    def match_order_daily(self, order):
        trade = Trade(order_id=order.order_id, exchange=order.exchange, ticker=order.ticker, asset_type=order.asset_type,
                      direction=order.direction, trade_volume=order.volume, source=order.source, time=self.day_data['time'])
        trade.trade_id = self.new_trade_id()
        # adjust trade price with trade fee and slippage 
        if order.direction == Direction.BUY:
            trade.trade_price = order.price * (1 + self.stock_slippage)

        elif order.direction == Direction.SELL:
            trade.trade_price = order.price * (1 - self.stock_slippage)
        return [trade]

    def switch_day(self):
        pass

    def terminate(self):
        pass

    def new_trade_id(self):
        self._cur_trade_id += 1
        return self._cur_trade_id


from .basic_data_part1 import BasicDataPart1
from .basic_data_part2_fund_ret import BasicFundRet
from .style_analysis_data import StockFactor
from .basic_data_helper import BasicDataHelper


class BasicDataProcessor(object):
    def __init__(self):
        self._data_helper = BasicDataHelper()
        self.basic_data_part1 = BasicDataPart1(self._data_helper)
        self.basic_fund_ret = BasicFundRet(self._data_helper)
        self._style_analysis = StockFactor(self._data_helper)

    def process_all(self, start_date, end_date):
        failed_tasks = []

        failed_tasks.extend(self.basic_data_part1.process_all(start_date, end_date))
        failed_tasks.extend(self.basic_fund_ret.process_all(end_date))
        failed_tasks.extend(self._style_analysis.process_all(start_date, end_date))

        return failed_tasks

    def get_updated_count(self):
        return self._data_helper._updated_count




from typing import List

from .fund_score_processor import FundScoreProcessor
from .fund_indicator_processor import FundIndicatorProcessor
from .derived_index_val import IndexValProcess
from .style_analysis_processor import StyleAnalysisProcessor
from .derived_data_helper import DerivedDataHelper
from .fund_indicator_processor_weekly import FundIndicatorProcessorWeekly
from .fund_indicator_processor_monthly import FundIndicatorProcessorMonthly
from .derived_index_ret_vol import IndexIndicatorProcessor
from .derived_fund_aic import AIPProcess
from ...api.basic import BasicDataApi


class DerivedDataProcessor:
    def __init__(self):
        self._data_helper = DerivedDataHelper()
        self.fund_indicator_processor = FundIndicatorProcessor(self._data_helper)
        self.fund_indicator_processor_weekly = FundIndicatorProcessorWeekly(self._data_helper)
        self.fund_indicator_processor_monthly = FundIndicatorProcessorMonthly(self._data_helper)
        self.fund_score_processor = FundScoreProcessor(self._data_helper)
        self.index_val_processor = IndexValProcess(self._data_helper)
        self.index_indicator = IndexIndicatorProcessor(self._data_helper)
        self.fund_aip = AIPProcess(self._data_helper)
        self.style_analysis_processors: List[StyleAnalysisProcessor] = []
        # 暂时只算这三个universe
        for universe in ('hs300', 'csi800', 'all'):
            sap = StyleAnalysisProcessor(self._data_helper, universe)
            self.style_analysis_processors.append(sap)

    def process_all(self, start_date, end_date):
        failed_tasks = []
        failed_tasks.extend(self.fund_indicator_processor.process(start_date, end_date))
        failed_tasks.extend(self.fund_score_processor.process(start_date, end_date))
        failed_tasks.extend(self.index_val_processor.process(start_date, end_date))
        failed_tasks.extend(self.index_indicator.process(end_date))
        failed_tasks.extend(self.fund_aip.process(start_date, end_date))
        for sap in self.style_analysis_processors:
            failed_tasks.extend(sap.process(start_date, end_date))

        # 获取下一个交易日
        api = BasicDataApi()
        trading_day_df = api.get_trading_day_list(start_date=end_date)
        if trading_day_df.shape[0] <= 1:
            print(f'get trading days start with {end_date} failed')
            failed_tasks.append('get_trading_day for weekly/monthly indicator')
        else:
            next_trading_day = trading_day_df.iloc[1, :].datetime
            print(f'got next trading day {next_trading_day}')
            failed_tasks.extend(self.fund_indicator_processor_weekly.process(start_date, end_date, next_trading_day))
            failed_tasks.extend(self.fund_indicator_processor_monthly.process(start_date, end_date, next_trading_day))
        return failed_tasks

    def get_updated_count(self):
        return self._data_helper._updated_count


if __name__ == '__main__':
    ddp = DerivedDataProcessor()
    start_date = '20200605'
    end_date = '20200605'
    # ddp.process_all(start_date, end_date)
    # ddp.fund_indicator_processor.process(start_date, end_date)
    # ddp.fund_score_processor.process(start_date, end_date)
    # import pandas as pd
    # date_list = pd.date_range(end='2010-05-31', periods=65, freq='M').sort_values(ascending=False).to_pydatetime()
    # for date in date_list:
    #     date = date.strftime('%Y%m%d')
    #     ddp.fund_indicator_processor_monthly.process(date, date)




from typing import List, Dict, Optional
import pandas as pd
import numpy as np
import traceback
import math
import json
import datetime
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler, MinMaxScaler

from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...view.derived_models import FundScoreExtended
from ..basic.basic_data_helper import BasicDataHelper
from .derived_data_helper import DerivedDataHelper


class FundScoreExtendedProcess:

    # 无风险收益率
    _RF = 0.011
    _TRADING_DAYS_PER_YEAR = 242
    _BRINSON_BEGIN_DATE = '20191231'
    _BRINSON_END_DATE = '20200630'
    _BRINSON_CLASS_LIST = ("混合型基金", "股票型基金")

    def __init__(self, data_helper: DerivedDataHelper):
        self._data_helper = data_helper
        self._raw_api = RawDataApi()
        self._basic_api = BasicDataApi()

    def init(self, start_date: str, end_date: str):
        start_date: str = (pd.to_datetime(start_date) - datetime.timedelta(days=365*3+1)).strftime('%Y%m%d')
        print(f'start date for calc: {start_date}')
        self._end_date: str = end_date

        # 获取区间内交易日列表
        all_trading_days: pd.Series = self._basic_api.get_trading_day_list().drop(columns='_update_time').set_index('datetime')
        self._trading_days: pd.Series = all_trading_days.loc[pd.to_datetime(start_date).date():pd.to_datetime(end_date).date()]

        # 基金净值数据、指数价格数据的index对齐到交易日列表
        fund_nav: pd.DataFrame = self._basic_api.get_fund_nav_with_date(start_date, end_date)
        self._fund_nav: pd.DataFrame = fund_nav.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').reindex(self._trading_days.index).fillna(method='ffill')
        index_price: pd.DataFrame = self._basic_api.get_index_price().drop(columns='_update_time')
        # 有的index最近没有数据，如活期存款利率/各种定期存款利率等，需要先reindex到全部的交易日列表上ffill
        index_price = index_price.pivot_table(index='datetime', columns='index_id', values='close').reindex(all_trading_days.index).fillna(method='ffill')
        # 再reindex到参与计算的交易日列表上
        self._index_price: pd.DataFrame = index_price.reindex(self._trading_days.index)
        pd.testing.assert_index_equal(self._fund_nav.index, self._index_price.index)

        try:
            # 这个指数有一天的点数是0，特殊处理一下
            self._index_price['spi_spa'] = self._index_price['spi_spa'].where(self._index_price.spi_spa != 0).fillna(method='ffill')
        except KeyError:
            pass

        # 对数净值取差分并且去掉第一个空值，得到基金对数收益率数列
        self._fund_ret: pd.DataFrame = np.log(self._fund_nav).diff().iloc[1:, :]

        # 计算benchmark return，且index对齐到fund_ret
        benchmark_ret: pd.DataFrame = self._get_benchmark_return()
        self._benchmark_ret: pd.DataFrame = benchmark_ret.reindex(self._fund_ret.index)
        pd.testing.assert_index_equal(self._fund_ret.index, self._benchmark_ret.index)

        # 获取待计算的基金列表
        fund_list: pd.DataFrame = self._basic_api.get_fund_info().drop(columns='_update_time')
        self._fund_list: pd.DataFrame = fund_list[fund_list.structure_type <= 1]
        # 获取wind一级分类
        self._wind_class_1: np.ndarray = fund_list.wind_class_1.unique()

    def _get_benchmark_return(self) -> pd.DataFrame:
        benchmark_list: Dict[str, float] = {}
        fund_benchmark_df: pd.DataFrame = self._basic_api.get_fund_benchmark().drop(columns='_update_time')
        # 遍历每一只基金的benchmark进行处理
        for row in fund_benchmark_df.itertuples(index=False):
            values: List[pd.Series] = []
            cons: float = 0
            # 空的benchmark表明我们没有对应的指数或无法解析公式
            if row.benchmark_s:
                benchmark: Dict[str, float] = json.loads(row.benchmark_s)
                benchmark_raw: Dict[str, float] = eval(row.benchmark)
                for (index, weight), index_raw in zip(benchmark.items(), benchmark_raw.keys()):
                    if index_raw == 1:
                        # 表示benchmark中该项为常数
                        cons += weight
                    elif index_raw.startswith('RA000'):
                        if weight == -1:
                            # 表示我们无法解析公式
                            print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                        else:
                            try:
                                ra: pd.Series = self._index_price.loc[:, index]
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                print(f'[benchmark_return] Error: Data Missing: {row.fund_id} {index} {index_raw}')
                            else:
                                values.append(ra.iloc[1:] * 0.01 * weight)
                    else:
                        if weight == -1:
                            # 表示我们无法解析公式
                            print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                        else:
                            try:
                                ra: pd.Series = self._index_price.loc[:, index]
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                print(f'Error: Data Missing: {row.fund_id} {index} {index_raw}')
                            else:
                                ra = np.log(ra).diff().iloc[1:]
                                values.append(ra * weight)
            if values or cons:
                the_sum: float = sum(values)
                if cons:
                    the_sum += np.log(math.pow(1 + cons, 1 / self._TRADING_DAYS_PER_YEAR))
                benchmark_list[row.fund_id] = the_sum
            else:
                benchmark_list[row.fund_id] = np.nan
        return pd.DataFrame.from_dict(benchmark_list)

    '''
    _get_resample_ret
    输入：日频率数据
    输出：月频率数据
    备注：输入需要DatetimeIndex
    '''
    @staticmethod
    def _get_resample_ret(df: pd.DataFrame) -> pd.DataFrame:
        df = df.set_axis(pd.to_datetime(df.index), inplace=False)
        return df.resample('1M').sum(min_count=1)

    @staticmethod
    def _from_rq_id(x: str) -> Optional[str]:
        _values = x.split('.')
        if _values[1] == 'XSHG':
            return _values[0] + '.SH'
        elif _values[1] == 'XSHE':
            return _values[0] + '.SZ'

    @staticmethod
    def _to_rq_id(x: str) -> Optional[str]:
        _values = x.split('.')
        if _values[1] == 'SH':
            return _values[0] + '.XSHG'
        elif _values[1] == 'SZ':
            return _values[0] + '.XSHE'

    @staticmethod
    def _brinson_calc_index(x: pd.DataFrame) -> pd.Series:
        total_weight: float = x.weight.sum()
        total_r: float = (x.weight * x.ret).sum()
        return pd.Series({'index_w': total_weight, 'index_r': total_r / total_weight})

    @staticmethod
    def _brinson_calc_fund(x: pd.DataFrame, index_datas: pd.DataFrame) -> pd.Series:
        total_weight: float = x.F_PRT_STKVALUE.sum()
        if total_weight != 0:
            total_r: float = (x.F_PRT_STKVALUE * x.ret).sum()
            fund_r = total_r / total_weight
        else:
            data = index_datas[index_datas.bl_sws_ind_code == x.bl_sws_ind_code]
            fund_r = data.index_r
        return pd.Series({'fund_w': total_weight, 'fund_r': fund_r})

    def _brinson_calc(self, fund_df: pd.DataFrame, index_datas: pd.DataFrame) -> pd.Series:
        # 处理一下基金持仓的市值权重
        stock_value_sum: float = fund_df.F_PRT_STKVALUE.sum()
        if stock_value_sum != 0:
            fund_df['F_PRT_STKVALUE'] /= stock_value_sum
        # 依次处理基金持仓个股所属的每一个申万一级行业
        summmary_list = fund_df.groupby('bl_sws_ind_code', sort=False).apply(FundScoreExtendedProcess._brinson_calc_fund, index_datas=index_datas)
        if summmary_list.empty:
            return
        # 结果和index数据放在一起
        summmary_list = summmary_list.join(index_datas, how='outer')
        summmary_list[['fund_w', 'fund_r']] = summmary_list.apply(lambda x: pd.Series([0, x.index_r]) if pd.isnull(x.fund_w) else pd.Series([x.fund_w, x.fund_r]), axis=1)
        # 计算分数
        summmary_list['R'] = summmary_list.loc[:, 'fund_w'] * summmary_list.loc[:, 'fund_r']
        summmary_list['Rs'] = summmary_list.loc[:, 'index_w'] * summmary_list.loc[:, 'fund_r']
        summmary_list['Bs'] = summmary_list.loc[:, 'fund_w'] * summmary_list.loc[:, 'index_r']
        summmary_list['B'] = summmary_list.loc[:, 'index_w'] * summmary_list.loc[:, 'index_r']
        summmary_list['Selection_BHB'] = summmary_list.loc[:, 'Rs'] - summmary_list.loc[:, 'B']
        summmary_list['Allocation_BHB'] = summmary_list.loc[:, 'Bs'] - summmary_list.loc[:, 'B']
        summmary_list['interaction_BHB'] = summmary_list.loc[:, 'R'] + summmary_list.loc[:, 'B'] - summmary_list.loc[:, 'Rs'] - summmary_list.loc[:, 'Bs']
        rd = summmary_list.loc[:, 'B'].sum()
        summmary_list['Allocation_BF'] = (summmary_list.loc[:, 'fund_w'] - summmary_list.loc[:, 'index_w']) * (summmary_list.loc[:, 'index_r'] - rd)
        return pd.Series({dim: summmary_list[dim].sum() for dim in ('Selection_BHB', 'Allocation_BHB', 'interaction_BHB')})

    def _brinson(self, start_date: str, end_date: str, benchmark: str = 'hs300'):
        index_weight_df: pd.DataFrame = self._raw_api.get_rq_index_weight([benchmark], start_date, end_date)
        # 暂时只用区间内第一天的数据
        index_weight: pd.Series = index_weight_df.iloc[0, :]
        # 将stock_list和weight_list打开
        index_whole_data = pd.DataFrame({'index_id': index_weight.at['index_id'], 'stock': json.loads(index_weight.at['stock_list']), 'weight': json.loads(index_weight.at['weight_list'])})
        # 米筐ID转换过来
        index_whole_data['stock'] = index_whole_data.stock.map(FundScoreExtendedProcess._from_rq_id)
        # index下的股票列表
        stock_list: List[str] = index_whole_data.stock.to_list()

        # blocklist = ['160218!0']
        # blocklist = ['001274!0','001428!0','002232!0','002834!0','002839!0','004049!0','004051!0', '005950!0','006050!0','006567!0','007234!0','007350!0','007501!0','007514!0','519618!0', '960009!0','960010!0','960013!0','960014!0','960015!0','960017!0','960019!0','960023!0', '960025!0', '960026!0', '960030!0', '960031!0', '960032!0', '960041!0', 'F050004!0', 'F050026!0', 'F080012!0', 'F161616!0', 'F202003!0', 'F450004!0', 'F450005!0', ]
        # brinson分类下的基金列表
        fund_list: List[str] = []
        for cl in self._BRINSON_CLASS_LIST:
            fund_list.extend(self._fund_list[self._fund_list.wind_class_1 == cl].wind_id.array)

        # 获取所有基金的股票持仓及持仓占比
        fund_stock_df: pd.DataFrame = self._raw_api.get_wind_fund_stock_portfolio(start_date, end_date, fund_list)
        # 暂时只用第一天的数据
        first_day = pd.to_datetime(fund_stock_df.F_PRT_ENDDATE).nsmallest(1).array[0].strftime('%Y%m%d')
        fund_whole_data: pd.DataFrame = fund_stock_df[fund_stock_df.F_PRT_ENDDATE == first_day]
        fund_whole_data = fund_whole_data[['S_INFO_WINDCODE', 'S_INFO_STOCKWINDCODE', 'F_PRT_STKVALUE']]

        # 计算全部股票列表
        stock_list.extend(fund_whole_data.S_INFO_STOCKWINDCODE.unique())
        stock_list_set = set(stock_list)

        # 获取全时间区间内所有这些股票的价格数据
        stock_list_set_rq = {FundScoreExtendedProcess._to_rq_id(one) for one in stock_list_set}
        stock_list_set_rq.discard(None)
        price_df: pd.DataFrame = self._basic_api.get_stock_price(stock_list_set_rq, start_date, end_date)
        price_df = price_df.pivot_table(index='datetime', columns='stock_id', values='close')
        # 米筐ID转换过来
        price_df = price_df.set_axis([FundScoreExtendedProcess._from_rq_id(one) for one in price_df.columns], axis=1, inplace=False)
        # 计算区间收益率
        ret: pd.Series = price_df.apply(lambda x: np.log(x.array[-1]) - np.log(x.array[0]))
        ret = ret.rename('ret')
        # 加到index上
        index_whole_data = index_whole_data.set_index('stock').join(ret, on='stock')

        # 获取各股票申万一级行业指数代码列表
        sw_ind_code: pd.Series = self._raw_api.get_em_stock_info(stock_list_set)
        sw_ind_code['bl_sws_ind_code'] = sw_ind_code.bl_sws_ind_code.transform(lambda x: x.split('-')[0])
        sw_ind_code = sw_ind_code[['bl_sws_ind_code', 'stock_id']].set_index('stock_id')
        # index加一列行业指数代码
        index_whole_data = index_whole_data.join(sw_ind_code, on='stock')

        # index的相关数据只需要计算一次即可，在这里算好
        index_datas: pd.DataFrame = index_whole_data.groupby(by='bl_sws_ind_code', sort=False).apply(FundScoreExtendedProcess._brinson_calc_index)
        # 依次计算每只基金
        result: pd.DataFrame = fund_whole_data.groupby(by='S_INFO_WINDCODE', sort=False).apply(lambda x: self._brinson_calc(x.join(ret, on='S_INFO_STOCKWINDCODE').join(sw_ind_code, on='S_INFO_STOCKWINDCODE'), index_datas))
        # 加上fund_id以及wind_class_1
        bdh = BasicDataHelper()
        result['fund_id'] = result.index.map(lambda x: bdh._get_fund_id_from_order_book_id(x.split('.')[0], pd.to_datetime(start_date).date()))
        result = result[result.fund_id.notna()]
        wind_class_dict: Dict[str] = self._fund_list[['fund_id', 'wind_class_1']].set_index('fund_id').to_dict()['wind_class_1']
        result['wind_class_1'] = result.fund_id.map(wind_class_dict)
        return result.set_index('fund_id')

    @staticmethod
    def _lambda_2(x: pd.Series, fund_ret_sampled: pd.DataFrame):
        temp: float = x.var()
        if pd.isnull(temp):
            return np.nan
        elif temp == 0:
            return 0

        fund_ret: pd.Series = fund_ret_sampled[x.name]
        if fund_ret.count() <= 1:
            return np.nan
        return x.cov(fund_ret) / temp

    @staticmethod
    def _lambda_1(x: pd.Series, fund_ret_sampled: pd.DataFrame):
        total = pd.concat({'Y': fund_ret_sampled[x.name], 'x': x}, axis=1)
        total = total[total.notna().all(axis=1)]
        if total.empty:
            return np.nan
        Y: pd.Series = total['Y']
        x = total['x']
        if x.count() != Y.count():
            return np.nan
        X: pd.DataFrame = pd.concat([x, x], axis=1)
        X.columns = [0, 1]
        X[0][X[0] < 0] = 0
        X[1][X[1] > 0] = 0
        regr = LinearRegression()
        regr.fit(X, Y)
        return regr.coef_[0] - regr.coef_[1]

    def calc(self):
        # 计算月度数据
        fund_ret_sampled: pd.DataFrame = FundScoreExtendedProcess._get_resample_ret(self._fund_ret)
        benchmark_ret_sampled: pd.DataFrame = FundScoreExtendedProcess._get_resample_ret(self._benchmark_ret)
        # 对齐基金收益和benchmark收益的月度数据的columns
        avail_fund_list: pd.Index = fund_ret_sampled.columns.intersection(benchmark_ret_sampled.columns)
        fund_ret_sampled = fund_ret_sampled.loc[:, avail_fund_list]
        benchmark_ret_sampled = benchmark_ret_sampled.loc[:, avail_fund_list]
        pd.testing.assert_index_equal(fund_ret_sampled.columns, benchmark_ret_sampled.columns)
        pd.testing.assert_index_equal(fund_ret_sampled.index, benchmark_ret_sampled.index)
        print('to calc basic factors')

        # beta
        beta = benchmark_ret_sampled.apply(self._lambda_2, fund_ret_sampled=fund_ret_sampled)

        # sharpe_ratio_M
        annualized_ret = fund_ret_sampled.sum(min_count=1) * 12 / fund_ret_sampled.shape[0]
        annualized_vol_0 = fund_ret_sampled.std(ddof=0) * math.sqrt(12)
        sharpe_ratio_M = (annualized_ret - self._RF) / annualized_vol_0

        # information_ratio
        ex_ret = fund_ret_sampled - benchmark_ret_sampled
        annualized_ex_ret = ex_ret.sum(min_count=1) * 12 / fund_ret_sampled.shape[0]
        annualized_vol_1 = fund_ret_sampled.std() * math.sqrt(12)
        information_ratio = annualized_ex_ret / annualized_vol_1

        # jensen_alpha
        rm = benchmark_ret_sampled.sum(min_count=1) * 12 / benchmark_ret_sampled.shape[0]
        jensen_alpha = annualized_ret - self._RF - beta * (rm - self._RF)

        # max_drawdown
        max_drawdown = 1 - (self._fund_nav / self._fund_nav.cummax()).min()

        # chang_lewellen
        chang_lewellen = benchmark_ret_sampled.apply(FundScoreExtendedProcess._lambda_1, fund_ret_sampled=fund_ret_sampled)

        # ret and vol
        ret = np.exp(self._fund_ret.sum(min_count=1)) - 1
        vol = self._fund_ret.std(ddof=0) * math.sqrt(12)

        total_df = pd.DataFrame.from_dict({'ret': ret, 'vol': vol, 'sr': sharpe_ratio_M, 'ir': information_ratio, 'beta': beta, 'alpha': jensen_alpha, 'mdd': max_drawdown, 'timing': chang_lewellen})
        total_df = total_df.replace([np.Inf, -np.Inf], np.nan)

        # brinson scores
        print('basic factors done, to calc brinson scores')
        brison_score_df = self._brinson(self._BRINSON_BEGIN_DATE, self._BRINSON_END_DATE)
        print('calc brinson scores done, to generate scores')

        scaler = StandardScaler()
        minmax_scaler = MinMaxScaler()
        # blocklist = ['001274!0','001428!0','002232!0','002834!0','002839!0','004049!0','004051!0', '005950!0','006050!0','006567!0','007234!0','007350!0','007501!0','007514!0','519618!0', '960009!0','960010!0','960013!0','960014!0','960015!0','960017!0','960019!0','960023!0', '960025!0', '960026!0', '960030!0', '960031!0', '960032!0', '960041!0', 'F050004!0', 'F050026!0', 'F080012!0', 'F161616!0', 'F202003!0', 'F450004!0', 'F450005!0', ]
        for cl in self._wind_class_1:
            # 与该类型下的基金列表取交集作为新的索引
            df = total_df.reindex(total_df.index.intersection(self._fund_list[self._fund_list.wind_class_1 == cl].fund_id))
            # 每列标准化后即为各列的score
            df_standardized = scaler.fit_transform(df)
            df = df.join(pd.DataFrame(df_standardized, index=df.index, columns=[one + '_score' for one in df.columns]))
            # 计算return score
            df['return_score'] = df['ret_score'].add(df['sr_score'], fill_value=0).add(df['ir_score'], fill_value=0).add(df['beta_score'], fill_value=0).add(df['alpha_score'], fill_value=0)
            # 计算robust score
            df['robust_score'] = df['mdd_score'].sub(df['vol_score'], fill_value=0)
            total_score_columns: List[str] = []
            for dim in ('return_score', 'robust_score', 'timing_score'):
                total_score_columns.append(dim)
                df[dim] = scaler.fit_transform(df[[dim]])
                # 百分制分数
                df[dim] = minmax_scaler.fit_transform(df[[dim]]) * 100
                df[dim.replace('score', 'rank')] = df[dim].rank(method='min', ascending=False)

            if cl in self._BRINSON_CLASS_LIST:
                df = df.join(brison_score_df[brison_score_df.wind_class_1 == cl], how='outer')
                for dim in ('Selection_BHB', 'Allocation_BHB', 'interaction_BHB'):
                    column_name = dim.split('_')[0].lower() + '_score'
                    total_score_columns.append(column_name)
                    df[column_name] = scaler.fit_transform(df[[dim]])
                    df = df.drop(columns=dim)
                    # 百分制分数
                    df[column_name] = minmax_scaler.fit_transform(df[[column_name]]) * 100
                    df[column_name.replace('score', 'rank')] = df[column_name].rank(method='min', ascending=False)

            # total_score = scaler.fit_transform(pd.DataFrame(df['return_score'].add(df['robust_score'], fill_value=0).add(df['timing_score'], fill_value=0)))
            total_score = pd.Series(index=df.index)
            for dim in total_score_columns:
                total_score = total_score.add(df[dim], fill_value=0)
            df['total_score'] = minmax_scaler.fit_transform(pd.DataFrame(total_score)) * 100
            df['total_rank'] = df['total_score'].rank(method='min', ascending=False)

            df['wind_class_1'] = cl
            df['datetime'] = self._end_date
            df = df.reset_index().rename(columns={'index': 'fund_id'})
            # df.to_csv('./res/' + cl + '_indicator.csv', encoding="utf_8_sig")
            print(f'calc score for funds in {cl} done')
            self._data_helper._upload_derived(df, FundScoreExtended.__table__.name)

    def process(self, start_date: str, end_date: str) -> List[str]:
        failed_tasks = []
        try:
            self.init(start_date, end_date)
            print('init done, begin to calc score')
            self.calc()
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_score_extended')
        return failed_tasks


if __name__ == '__main__':
    start_date = '20200730'
    end_date = '20200730'
    fse = FundScoreExtendedProcess(DerivedDataHelper())
    fse.process(start_date, end_date)



from sqlalchemy import CHAR, Column, Integer, Index, BOOLEAN, text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.mysql import DOUBLE, DATE, DATETIME


class Base():
    _update_time = Column('_update_time', DATETIME, nullable=False, server_default=text('CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP'))  # 更新时间


# make this column at the end of every derived table
Base._update_time._creation_order = 9999
Base = declarative_base(cls=Base)


class IndexVolatility(Base):
    '''指数波动率'''

    __tablename__ = 'index_volatility'

    index_id = Column(CHAR(20), primary_key=True) # 指数ID
    datetime = Column(DATE, primary_key=True) # 日期
    w1_vol = Column(DOUBLE(asdecimal=False)) # 近一周波动率
    m1_vol = Column(DOUBLE(asdecimal=False)) # 近一月波动率
    m3_vol = Column(DOUBLE(asdecimal=False)) # 近三月波动率
    m6_vol = Column(DOUBLE(asdecimal=False)) # 近半年波动率
    y1_vol = Column(DOUBLE(asdecimal=False)) # 近一年波动率
    y3_vol = Column(DOUBLE(asdecimal=False)) # 近三年波动率
    y5_vol = Column(DOUBLE(asdecimal=False)) # 近五年波动率
    y10_vol = Column(DOUBLE(asdecimal=False)) # 近十年波动率
    this_y_vol = Column(DOUBLE(asdecimal=False)) # 今年以来波动率
    this_s_vol = Column(DOUBLE(asdecimal=False)) # 本季度以来波动率
    cumulative_vol = Column(DOUBLE(asdecimal=False)) # 成立至今波动率

    __table_args__ = (
        Index('idx_index_volatility_datetime', 'datetime'),
    )


class IndexReturn(Base):
    '''指数收益率'''

    __tablename__ = 'index_return'

    index_id = Column(CHAR(20), primary_key=True) # 指数ID
    datetime = Column(DATE, primary_key=True) # 日期
    w1_ret = Column(DOUBLE(asdecimal=False)) # 近一周收益率
    m1_ret = Column(DOUBLE(asdecimal=False)) # 近一月收益率
    m3_ret = Column(DOUBLE(asdecimal=False)) # 近三月收益率
    m6_ret = Column(DOUBLE(asdecimal=False)) # 近半年收益率
    y1_ret = Column(DOUBLE(asdecimal=False)) # 近一年收益率
    y3_ret = Column(DOUBLE(asdecimal=False)) # 近三年收益率
    y5_ret = Column(DOUBLE(asdecimal=False)) # 近五年收益率
    y10_ret = Column(DOUBLE(asdecimal=False)) # 近十年收益率
    this_y_ret = Column(DOUBLE(asdecimal=False)) # 今年以来收益率
    this_s_ret = Column(DOUBLE(asdecimal=False)) # 本季度以来收益率
    cumulative_ret = Column(DOUBLE(asdecimal=False)) # 成立至今收益率

    __table_args__ = (
        Index('idx_index_return_datetime', 'datetime'),
    )


class FundAlpha(Base):
    '''基金超额收益'''

    __tablename__ = 'fund_alpha'

    track_err = Column(DOUBLE(asdecimal=False)) # 跟踪误差
    this_y_alpha = Column(DOUBLE(asdecimal=False)) # 今年以来超额收益
    cumulative_alpha = Column(DOUBLE(asdecimal=False)) # 成立以来超额收益
    w1_alpha = Column(DOUBLE(asdecimal=False)) # 近一周收益率
    m1_alpha = Column(DOUBLE(asdecimal=False)) # 近一月超额收益
    m3_alpha = Column(DOUBLE(asdecimal=False)) # 近三月超额收益
    m6_alpha = Column(DOUBLE(asdecimal=False)) # 近半年超额收益
    y1_alpha = Column(DOUBLE(asdecimal=False)) # 近一年超额收益
    y3_alpha = Column(DOUBLE(asdecimal=False)) # 近三年超额收益
    y5_alpha = Column(DOUBLE(asdecimal=False)) # 近五年超额收益
    y10_alpha = Column(DOUBLE(asdecimal=False)) # 近十年超额收益
    fund_id = Column(CHAR(16), primary_key=True) # 基金ID
    datetime = Column(DATE, primary_key=True) # 日期

    __table_args__ = (
        Index('idx_fund_alpha_datetime', 'datetime'),
    )

class IndexValuation(Base):
    '''指数估值'''

    __tablename__ = 'index_valuation'
    index_id = Column(CHAR(20), primary_key=True) #指数ID
    datetime = Column(DATE, primary_key=True) # 日期

    pb_mrq = Column(DOUBLE(asdecimal=False)) # 市净率-MRQ
    pe_ttm = Column(DOUBLE(asdecimal=False)) # 市盈率-MMT
    roe = Column(DOUBLE(asdecimal=False)) # 净资产收益率
    ps_ttm = Column(DOUBLE(asdecimal=False)) # 市销率—MMT
    dy = Column(DOUBLE(asdecimal=False)) # 股息率
    pcf_ttm = Column(DOUBLE(asdecimal=False)) # 市现率-MMT
    peg_ttm = Column(DOUBLE(asdecimal=False)) # 市盈率相对盈利增长比率-MMT
    pe_pct = Column(DOUBLE(asdecimal=False)) # PE百分位
    pb_pct = Column(DOUBLE(asdecimal=False)) # PB百分位
    ps_pct = Column(DOUBLE(asdecimal=False)) # PS百分位
    val_score = Column(DOUBLE(asdecimal=False)) # 估值评分
    eps_ttm = Column(DOUBLE(asdecimal=False)) # 每股收益—MMT
    roe_pct = Column(DOUBLE(asdecimal=False)) # ROE百分位
    est_peg = Column(DOUBLE(asdecimal=False)) # 预测peg

    __table_args__ = (
        Index('idx_index_valuation_develop_datetime', 'datetime'),
    )


class FundIndicator(Base):
    '''基金评价指标'''

    __tablename__ = 'fund_indicator'
    fund_id = Column(CHAR(10), primary_key=True) # 基金ID
    datetime = Column(DATE, primary_key=True) # 日期
    beta = Column(DOUBLE(asdecimal=False)) # 风险指数
    alpha = Column(DOUBLE(asdecimal=False)) # 投资回报
    track_err = Column(DOUBLE(asdecimal=False)) # 跟踪误差
    timespan = Column(DOUBLE(asdecimal=False)) # 历史数据跨度(年)
    fee_rate = Column(DOUBLE(asdecimal=False)) # 费率
    info_ratio = Column(DOUBLE(asdecimal=False)) # 信息比率
    treynor = Column(DOUBLE(asdecimal=False)) # 特雷诺比率
    mdd = Column(DOUBLE(asdecimal=False)) #净值最大回撤
    down_risk = Column(DOUBLE(asdecimal=False)) #下行风险
    ret_over_period = Column(DOUBLE(asdecimal=False)) #区间收益率
    annual_avg_daily_ret = Column(DOUBLE(asdecimal=False)) #年化日均收益
    annual_vol = Column(DOUBLE(asdecimal=False)) #年化波动率
    annual_ret = Column(DOUBLE(asdecimal=False)) #年化收益
    m_square = Column(DOUBLE(asdecimal=False)) #M平方测度 风险调整收益指标
    time_ret = Column(DOUBLE(asdecimal=False)) #择时收益
    var = Column(DOUBLE(asdecimal=False)) #资产在险值
    r_square = Column(DOUBLE(asdecimal=False)) #决定系数R方
    sharpe = Column(DOUBLE(asdecimal=False)) #夏普率
    year_length = Column(DOUBLE(asdecimal=False)) #成立年限

    __table_args__ = (
        Index('idx_fund_indicator_datetime', 'datetime'),
    )


class FundIndicatorWeekly(Base):
    '''基金评价指标'''

    __tablename__ = 'fund_indicator_weekly'
    fund_id = Column(CHAR(10), primary_key=True) # 基金ID
    datetime = Column(DATE, primary_key=True) # 日期
    beta_w = Column(DOUBLE(asdecimal=False)) # 风险指数
    alpha_w = Column(DOUBLE(asdecimal=False)) # 投资回报
    track_err_w = Column(DOUBLE(asdecimal=False)) # 跟踪误差

    __table_args__ = (
        Index('idx_fund_indicator_datetime_weekly', 'datetime'),
    )

class FundIndicatorMonthly(Base):
    '''基金评价指标（月度）'''

    __tablename__ = 'fund_indicator_monthly'

    fund_id = Column(CHAR(10), primary_key=True)  # 基金ID
    datetime = Column(DATE, primary_key=True)  # 日期
    beta_m = Column(DOUBLE(asdecimal=False), nullable=False)  # beta
    sharpe_ratio_m = Column(DOUBLE(asdecimal=False), nullable=False)  # 夏普比率
    treynor_ratio_m = Column(DOUBLE(asdecimal=False), nullable=False)  # 特雷诺比率
    information_ratio_m = Column(DOUBLE(asdecimal=False), nullable=False)  # 信息比率
    jensen_alpha_m = Column(DOUBLE(asdecimal=False), nullable=False)  # 詹森指数
    calmar_ratio_m = Column(DOUBLE(asdecimal=False), nullable=False)  # 卡玛比率

    __table_args__ = (
        Index('idx_fund_indicator_monthly_datetime_monthly', 'datetime'),
    )


class FundIndicatorAnnual(Base):
    '''基金年度指标'''

    __tablename__ = 'fund_indicator_annual'

    fund_id = Column(CHAR(16), primary_key=True)  # 基金ID
    datetime = Column(DATE, primary_key=True)  # 日期
    beta_annual = Column(DOUBLE(asdecimal=False))  # beta
    alpha_annual = Column(DOUBLE(asdecimal=False))  # alpha
    excess_return_annual = Column(DOUBLE(asdecimal=False))  # 超额收益

    __table_args__ = (
        Index('idx_fund_indicator_datetime_annual', 'datetime'),
    )


class FundScore(Base):
    '''基金评分'''

    __tablename__ = 'fund_score'
    id = Column(Integer, primary_key=True)

    fund_id = Column(CHAR(10)) # 基金ID
    datetime = Column(DATE) # 日期
    score = Column(DOUBLE(asdecimal=False)) # 评分
    tag_name = Column(CHAR(64)) # 基金类别
    tag_type = Column(Integer) # 1 '大类资产评分'
    tag_method = Column(CHAR(40)) # 打标签版本号
    is_full = Column(BOOLEAN) # 1 true
    score_method = Column(Integer)
    desc_name = Column(CHAR(64))

    __table_args__ = (
        Index('idx_fund_score_datetime', 'datetime'),
    )


class FundScoreExtended(Base):
    '''基金评分(新)'''

    __tablename__ = 'fund_score_extended'

    fund_id = Column(CHAR(16), primary_key=True)  # 基金代码
    datetime = Column(DATE, primary_key=True)  # 日期
    wind_class_1 = Column(CHAR(64))  # Wind基金类型
    ret = Column(DOUBLE(asdecimal=False))  # ret
    vol = Column(DOUBLE(asdecimal=False))  # volatility
    sr = Column(DOUBLE(asdecimal=False))  # sharpe ratio
    ir = Column(DOUBLE(asdecimal=False))  # information ratio
    beta = Column(DOUBLE(asdecimal=False))  # beta
    alpha = Column(DOUBLE(asdecimal=False))  # alpha
    mdd = Column(DOUBLE(asdecimal=False))  # max drawdown
    timing = Column(DOUBLE(asdecimal=False))  # timing
    ret_score = Column(DOUBLE(asdecimal=False))  # score of ret
    vol_score = Column(DOUBLE(asdecimal=False))  # score of volatility
    sr_score = Column(DOUBLE(asdecimal=False))  # score of sharpe ratio
    ir_score = Column(DOUBLE(asdecimal=False))  # score of information ratio
    beta_score = Column(DOUBLE(asdecimal=False))  # score of beta
    alpha_score = Column(DOUBLE(asdecimal=False))  # score of alpha
    mdd_score = Column(DOUBLE(asdecimal=False))  # score of max drawdown
    return_score = Column(DOUBLE(asdecimal=False))  # score of return
    robust_score = Column(DOUBLE(asdecimal=False))  # score of robust
    timing_score = Column(DOUBLE(asdecimal=False))  # score of timing
    selection_score = Column(DOUBLE(asdecimal=False))  # score of selection
    allocation_score = Column(DOUBLE(asdecimal=False))  # score of allocation
    interaction_score = Column(DOUBLE(asdecimal=False))  # score of interaction
    return_rank = Column(DOUBLE(asdecimal=False))  # rank of return
    robust_rank = Column(DOUBLE(asdecimal=False))  # rank of robust
    timing_rank = Column(DOUBLE(asdecimal=False))  # rank of timing
    selection_rank = Column(DOUBLE(asdecimal=False))  # rank of selection
    allocation_rank = Column(DOUBLE(asdecimal=False))  # rank of allocation
    interaction_rank = Column(DOUBLE(asdecimal=False))  # rank of interaction
    total_score = Column(DOUBLE(asdecimal=False), nullable=False)  # score of total
    total_rank = Column(DOUBLE(asdecimal=False), nullable=False)  # rank of total

    __table_args__ = (
        Index('idx_fund_score_extended_datetime', 'datetime'),
    )


class AssetAllocationInfo(Base):
    '''十档大类资产组合'''

    __tablename__ = 'asset_allocation_info'
    allocation_id = Column(Integer, primary_key=True) # 组合id
    version = Column(Integer, primary_key=True) # 版本号

    hs300 =  Column(DOUBLE(asdecimal=False)) # 沪深300权重
    csi500 =  Column(DOUBLE(asdecimal=False)) # 中证500权重
    gem =  Column(DOUBLE(asdecimal=False)) # 创业板权重
    sp500rmb =  Column(DOUBLE(asdecimal=False)) # 标普500人民币权重
    national_debt = Column(DOUBLE(asdecimal=False)) # 国债权重
    gold =  Column(DOUBLE(asdecimal=False)) # 黄金权重
    credit_debt =  Column(DOUBLE(asdecimal=False)) # 信用债权重
    dax30rmb = Column(DOUBLE(asdecimal=False)) # 德国dax30权重
    real_state = Column(DOUBLE(asdecimal=False)) # 房地产权重 拼错了
    oil = Column(DOUBLE(asdecimal=False)) # 原油权重
    n225rmb = Column(DOUBLE(asdecimal=False)) # 日经225权重
    cash = Column(DOUBLE(asdecimal=False)) # 现金权重
    mdd = Column(DOUBLE(asdecimal=False)) # 最大回撤
    annual_ret = Column(DOUBLE(asdecimal=False)) # 年化收益
    sharpe = Column(DOUBLE(asdecimal=False)) # 夏普率
    recent_y5_ret = Column(DOUBLE(asdecimal=False)) # 最近五年收益
    annual_vol = Column(DOUBLE(asdecimal=False)) # 年化波动率
    mdd_d1 = Column(DATE) # 最大回撤开始时间
    mdd_d2 = Column(DATE) # 最大回撤结束时间
    start_date = Column(DATE) # 回测开始时间
    end_date = Column(DATE) # 回测结束时间

class StyleAnalysisFactorReturn(Base):
    '''风格分析因子收益'''

    __tablename__ = 'style_analysis_factor_return'

    universe_index = Column(CHAR(20), primary_key=True)  # universe, 全市场股票时值为all
    datetime = Column(DATE, primary_key=True)  # 日期
    latest_size = Column(DOUBLE(asdecimal=False))  # 规模
    bp = Column(DOUBLE(asdecimal=False))  # 价值
    short_term_momentum = Column(DOUBLE(asdecimal=False))  # 短期动量
    long_term_momentum = Column(DOUBLE(asdecimal=False))  # 长期动量
    high_low = Column(DOUBLE(asdecimal=False))  # 波动率
    const = Column(DOUBLE(asdecimal=False))  # 常数项

class BarraCNE5FactorReturn(Base):
    '''Barra CNE5因子收益'''

    __tablename__ = 'barra_cne5_factor_return'

    datetime = Column(DATE, primary_key=True)  # 日期
    market = Column(DOUBLE(asdecimal=False))
    leverage = Column(DOUBLE(asdecimal=False))
    cap = Column(DOUBLE(asdecimal=False))
    momentum = Column(DOUBLE(asdecimal=False))
    bp = Column(DOUBLE(asdecimal=False))
    earning = Column(DOUBLE(asdecimal=False))
    liquidity = Column(DOUBLE(asdecimal=False))
    growth = Column(DOUBLE(asdecimal=False))
    vol = Column(DOUBLE(asdecimal=False))

class AllocationDistribution(Base):
    '''有效前沿下各mdd最优策略'''

    __tablename__ = 'allocation_distribution'

    ef_id = Column(CHAR(8), primary_key=True)  # 有效前沿id
    version = Column(Integer, primary_key=True) # 版本号
    annual_ret = Column(DOUBLE(asdecimal=False)) #年化收益
    csi500 = Column(DOUBLE(asdecimal=False)) # 中证500权重
    gem = Column(DOUBLE(asdecimal=False)) #创业板权重
    gold = Column(DOUBLE(asdecimal=False)) #黄金权重
    hs300 = Column(DOUBLE(asdecimal=False)) #沪深300权重
    mdd = Column(DOUBLE(asdecimal=False)) # 最大回撤
    mdd_up_limit = Column(DOUBLE(asdecimal=False)) #最大回测档位上限
    mmf = Column(DOUBLE(asdecimal=False)) # 货币基金权重
    national_debt = Column(DOUBLE(asdecimal=False)) #国债权重
    sp500rmb = Column(DOUBLE(asdecimal=False)) #标普500权重
    cash = Column(DOUBLE(asdecimal=False)) #现金权重

class FundAIC(Base):
    '''基金定投收益收集'''

    __tablename__ = 'fund_aic'

    fund_id = Column(CHAR(20), primary_key=True)  # 基金ID
    datetime = Column(CHAR(20), primary_key=True)  # 时间
    y1_ret = Column(DOUBLE(asdecimal=False))  # 普通定投一年收益
    y3_ret = Column(DOUBLE(asdecimal=False))  # 普通定投三年收益
    y5_ret = Column(DOUBLE(asdecimal=False))  # 普通定投五年收益
    intel_y1_ret = Column(DOUBLE(asdecimal=False))  # 智能定投一年收益
    intel_y3_ret = Column(DOUBLE(asdecimal=False))  # 智能定投三年收益
    intel_y5_ret = Column(DOUBLE(asdecimal=False))  # 智能定投五年收益




import pandas as pd
import numpy as np
import datetime
import traceback
from functools import partial
from sklearn.metrics import r2_score
import time
import concurrent.futures
from ...manager.manager_fund import FundDataManager
from ...manager.score import FundScoreManager
from ...wrapper.mysql import DerivedDatabaseConnector
from ...view.derived_models import *


class FundIndicatorProcessorWeekly(object):

    TRADING_DAYS_PER_YEAR = 242
    NATURAL_DAYS_PER_YEAR = 365
    SHORT_TIME_SPAN = TRADING_DAYS_PER_YEAR
    LONG_TIME_SPAN = TRADING_DAYS_PER_YEAR * 3
    LONG_TERM_INDEX = ['hs300', 'csi500', 'mmf']
    MIN_TIME_SPAN = 10 #最小回归单位
    RESAMPLE_WORK_DAY_DIC = {'1W':5}

    def __init__(self, data_helper):
        self._data_helper = data_helper

    def init(self, start_date='20040101', end_date='20200626', dm=None, resample_day='1W'):
        if dm is None:
            self._dm = FundDataManager(start_time=start_date, end_time=end_date, score_manager=FundScoreManager())
            self._dm.init(score_pre_calc=False)
        else:
            self._dm = dm
        self.start_date = self._dm.start_date
        self.end_date = self._dm.end_date
        self.resample_day = resample_day

        # index price, fund nav 日期对齐至交易日
        self.trading_days = self._dm.dts.trading_days.datetime
        self.fund_nav = self._dm.dts.fund_nav.copy()
        self.fund_info = self._dm.dts.fund_info.copy()
        self.fund_to_index_dict = self.fund_info[['fund_id', 'index_id']].set_index('fund_id').to_dict()['index_id']
        self.index_price = self._dm.dts.index_price.copy()
        self.fund_nav = self.fund_nav.reindex(self.trading_days).loc[:self.end_date]
        self.index_price = self.index_price.reindex(self.trading_days)[:self.end_date]
        l1 = self.fund_nav.index.to_list().sort()
        l2 = self.index_price.index.to_list().sort()
        assert l1 == l2, f'date index of index price and fund nav are not identical, l1 {l1} lens {len(l1)}, l2 {l2} lens {len(l2)}'

        # 超过基金终止日的基金净值赋空
        fund_to_enddate_dict = self.fund_info[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']
        for fund_id in self.fund_nav.columns:
            fund_end_date = fund_to_enddate_dict[fund_id]
            if self.end_date > fund_end_date:
                self.fund_nav.loc[fund_end_date:,fund_id] = np.nan

        # resample and datetime convert
        self.fund_ret = np.log(self.fund_nav / self.fund_nav.shift(1))
        self.index_ret = np.log(self.index_price / self.index_price.shift(1))
        self.fund_ret = self.df_resample(self.fund_ret)
        self.index_ret = self.df_resample(self.index_ret)

        self.fund_ret = self.fund_ret.stack().reset_index(level=[0,1]).rename(columns={0:'ret'})
        self.fund_ret['index_id'] = self.fund_ret.fund_id.apply(lambda x: self.fund_to_index_dict[x])
        self.fund_ret = self.fund_ret.pivot_table(index = ['index_id','datetime'],columns='fund_id',values='ret')

        self.index_list = self.fund_ret.index.levels[0]
        self.index_fund = { index_id : [fund_idx for fund_idx, index_idx in self.fund_to_index_dict.items() if index_idx == index_id] for index_id in self.index_list}
        self.start_date_dic = self.fund_info[['fund_id','start_date']].set_index('fund_id').to_dict()['start_date']

    def df_resample(self, df):
        df.index = pd.to_datetime(df.index)
        df = df.resample(self.resample_day).sum()
        df.index = pd.Series(df.index).dt.date
        return df

    def get_time_range(self, index_id):
        if index_id in self.LONG_TERM_INDEX:
            return self.LONG_TIME_SPAN
        else:
            return self.SHORT_TIME_SPAN

    def _rolling_alpha_beta_time_ret_r2(self, x, res, df):
        # 回归相关的都在这里
        df_i = df.loc[x[0]:x[-1],]
        return self._rolling_alpha_beta_time_ret_r2_base(res,df_i)

    def _rolling_alpha_beta_time_ret_r2_base(self, res, df_i):
        if sum(df_i.fund) == 0:
            res.append({'alpha_w':np.Inf,'beta_w':np.Inf})
            return 1
        else:
            ploy_res = np.polyfit(y=df_i.fund, x=df_i.benchmark,deg=1)
            p = np.poly1d(ploy_res)
            beta = ploy_res[0]
            # 根据resample 日子做回归  比如当前是 1w 对应 5天
            alpha = ploy_res[1] * self.TRADING_DAYS_PER_YEAR / self.RESAMPLE_WORK_DAY_DIC[self.resample_day]
            res.append({'alpha_w': alpha,
                        'beta_w': beta})
            return 1

    def _process_fund_indicator_item(self, fund_id, time_range, fund_ret, index_id, index_ret):
        df = fund_ret[[fund_id]].join(index_ret).dropna()
        b_day = self.fund_nav[[fund_id]].dropna().index[0]
        df = df.loc[b_day:].rename(columns={index_id:'benchmark',fund_id:'fund'}).reset_index()
        res = []
        pd.Series(df.index).rolling(
            window=time_range,min_periods=self.MIN_TIME_SPAN).apply(
                partial(self._rolling_alpha_beta_time_ret_r2, res=res, df=df), raw=True)
        df = df.set_index('datetime')
        df = df.join(pd.DataFrame(res,index=df.index[-len(res):]))
        if 'alpha_w' not in df.columns:
            return None
        df['track_err_w'] = (df.fund - df.benchmark).rolling(window=time_range, min_periods=self.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR / self.RESAMPLE_WORK_DAY_DIC[self.resample_day])
        df['fund_id'] = fund_id
        df = df.drop(['fund','benchmark'], axis=1)
        df = df.replace({-np.Inf:None,np.Inf:None}).dropna(subset=['beta_w', 'alpha_w','track_err_w']).reset_index()
        if df.empty:
            return None
        return df

    def calculate_history(self, index_id_list:list=None):
        result = []
        index_list = self.index_list if index_id_list is None else index_id_list
        for index_id in index_list:
            _tm_1 = time.time()
            time_range = int(self.get_time_range(index_id) / self.RESAMPLE_WORK_DAY_DIC[self.resample_day])
            fund_list = self.index_fund[index_id]
            fund_list_calculate = self.fund_ret.loc[index_id].columns
            fund_list = list(set(fund_list) & set(fund_list_calculate))
            fund_ret = self.fund_ret.loc[index_id][fund_list].copy()
            index_ret = self.index_ret[[index_id]]
            for fund_id in fund_list:
                df_i = self._process_fund_indicator_item(fund_id, time_range, fund_ret, index_id, index_ret)
                result.append(df_i)
            _tm_2 = time.time()
            print(f'index {index_id} fund number : {len(fund_list)} finish, cost time {_tm_2 - _tm_1} s')
        result = [_ for _ in result if not _ is None]
        self.result = pd.concat(result, axis=0, sort=True)

    def save_db(self):
        self._data_helper._upload_derived(self.result, FundIndicatorWeekly.__table__.name)

    def process(self, start_date, end_date, next_trading_day):
        # 周更因子, resample出来的日期都是周日，在周日前的最后一个交易日更新
        failed_tasks = []
        try:
            end_date_dt = pd.to_datetime(end_date, infer_datetime_format=True).date()
            end_date_weekday = end_date_dt.weekday()
            next_trading_dt = pd.to_datetime(next_trading_day, infer_datetime_format=True).date()
            next_trading_weekday = next_trading_dt.weekday()
            if end_date_weekday < next_trading_weekday and next_trading_dt < end_date_dt + datetime.timedelta(weeks=1):
                # 表明本周后边还有交易日，今天不需要更新
                print(f'weekly indicator only update on the last day of week, not today {end_date_dt}')
                return failed_tasks
            else:
                print(f'weekly indicator update on the last day of week {end_date_dt}')
                start_date_dt = pd.to_datetime(start_date, infer_datetime_format=True).date()
                start_date = start_date_dt - datetime.timedelta(days=1150)  # 3年历史保险起见，多取几天 3*365=1095
                start_date = datetime.datetime.strftime(start_date, '%Y%m%d')
                self.init(start_date=start_date, end_date=end_date)
                self.calculate_history()
                end_date_dt = end_date_dt + datetime.timedelta(days=2)
                df = self.result[self.result['datetime'] == end_date_dt]
                self._data_helper._upload_derived(df, FundIndicatorWeekly.__table__.name)

        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_indicator_weekly')
        return failed_tasks



import pandas as pd
import numpy as np
import platform
import matplotlib as mpl
import pylab as pl
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
import matplotlib.dates as mdates
import datetime
from matplotlib.patches import Rectangle
from ....data.manager.score import FundScoreManager
CURRENT_PLATFORM = platform.system()
if CURRENT_PLATFORM == 'Darwin':
    mpl.rcParams['font.family'] = ['Heiti TC']
else:
    mpl.rcParams['font.family'] = ['STKaiti']
pd.plotting.register_matplotlib_converters()
class FundPainter(object):

    @staticmethod
    def plot_fund_weights(  fund_weights_history:dict, 
                            fund_cash_history:dict, 
                            fund_marktet_price_history:dict,
                            saa:dict):
        res = []
        for k,v in fund_weights_history.items():
            v = v
            v['date'] = k
            v['cash'] = fund_cash_history[k]  / fund_marktet_price_history[k]
            res.append(v)
        weights_df = pd.DataFrame(res).set_index('date')
        weights_df = weights_df.div(weights_df.sum(axis=1), axis=0)
        weights_df.fillna(0)[1:].plot.area(figsize=(18,9),legend=False,fontsize = 17)
        s = pl.title('fund weights history', fontsize=20)
        st = [k+':'+str(round(v,2)) for k, v in saa.items() if v > 0]
        st = ' '.join(st)
        plt.suptitle(st,y=0.87,fontsize=17)
        plt.grid()

    @staticmethod
    def plot_index_fund_fee( index_fee:pd.DataFrame):
        mpl.rcParams['font.size'] = 15
        p = index_fee.plot.pie(y='amount',figsize = (8,8))
        l = plt.legend(fontsize=17 ,loc = 'lower left')
        t = plt.title('index fee amount', fontsize=20)
        y = plt.ylabel(ylabel='fee amount', fontsize=17)
        plt.grid()

    @staticmethod
    def plot_fund_mdd_periods(  fund_mv:pd.DataFrame, 
                                fund_weights_history:dict, 
                                fund_nav:pd.DataFrame, 
                                fund_info:pd.DataFrame):
        df = fund_mv
        mdd_part1 = (df.loc[:, 'mv'] / df.loc[:, 'mv'].rolling(10000, min_periods=1).max())
        mdd = round(1 - mdd_part1.min(),4)
        mdd_date1 = df.loc[:mdd_part1.idxmin(),'mv'].idxmax()
        mdd_date2 = mdd_part1.idxmin()
        d1 = mdd_date1.strftime('%Y-%m-%d')
        d2 = mdd_date2.strftime('%Y-%m-%d')
        date_list = np.array(list(fund_weights_history.keys()))
        date_list = date_list[(date_list >= mdd_date1) & (date_list <= mdd_date2)]
        d = date_list[0]
        fund_list = [k for k,v in fund_weights_history[d].items() if isinstance(v, float) and (round(v,3) > 0)] 
        if 'cash' in fund_list:
            fund_list.remove('cash')
        df = fund_nav[fund_list].loc[mdd_date1:mdd_date2,:] 
        if df.empty:
            return 
        fig, ax = plt.subplots(figsize= [18,9])
        df = df/df.iloc[0]
        for col in df:
            df[col] = 1 - (1 - df[col].values )* fund_weights_history[d][col]
        table_df = fund_info.set_index('fund_id').loc[fund_list,:].reset_index()[['fund_id','desc_name','index_id']].sort_values(['index_id'])
        fund_desc_dict = fund_info.reset_index()[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']
        df_mdd = pd.DataFrame(((1-df.iloc[-1])*100).round(2))
        df_mdd.columns = ['max draw down %']
        table_df = table_df.set_index('fund_id').join(df_mdd).reset_index()
        fund_list = table_df.fund_id.tolist()
        df = df.rename(columns = fund_desc_dict)
        for col in df.columns:
            plt.plot(df.index, df[[col]], linewidth=1.0, label = col)
        plt.legend(fontsize=17 ,loc = 'lower left')
        plt.title('fund bt during mdd period  nav of all funds', fontsize=20)
        ax.xaxis.set_ticks_position('top')
        t = plt.table(cellText=table_df.values.tolist(),
          colLabels=table_df.columns,
          colWidths= [0.25,0.45,0.15,0.15],  
          loc='bottom',
          )
        t.auto_set_font_size(False)
        t.set_fontsize(17)
        t.auto_set_column_width('fund_id')
        t.AXESPAD = 0.1
        t.scale(1, 2)
        plt.grid()
        plt.suptitle(f'mdd : {mdd}, from {d1} to {d2}',y=0.87,fontsize=17)    
        plt.show()

    @staticmethod
    def plot_fund_mdd_amounts(  fund_mv:pd.DataFrame, 
                                fund_nav:pd.DataFrame, 
                                fund_info:pd.DataFrame,
                                fund_position_history:dict):
        df = fund_mv
        date_list = df.index.tolist()
        fund_ret = fund_nav / fund_nav.shift(1) - 1
        mdd_part1 = (df.loc[:, 'mv'] / df.loc[:, 'mv'].rolling(10000, min_periods=1).max())
        mdd_date1 = df.loc[:mdd_part1.idxmin(),'mv'].idxmax()
        mdd_date2 = mdd_part1.idxmin()
        res = []
        for d in fund_position_history:
            if d < mdd_date1 or d > mdd_date2:
                continue
            res_d = {}
            res_d['datetime'] = d
            if fund_position_history[d] == {}:
                continue
            else:
                for fund_id in fund_position_history[d]:
                    if fund_position_history[d][fund_id]['volume'] > 0:
                        res_d[fund_id] = fund_position_history[d][fund_id]['volume']
            res.append(res_d)
        volume_df = pd.DataFrame(res).set_index('datetime').loc[mdd_date1:mdd_date2]
        res = []
        for d, r in volume_df.iterrows():
            next_d = date_list[date_list.index(d) + 1]
            pos_n_1 = r.dropna()
            pos_n_1.name = 'pos'
            fund_id_list = pos_n_1.index.tolist()
            nav_n_1 = fund_nav.loc[d][fund_id_list]
            nav_n_1.name = 'nav'
            fund_ret_n = fund_ret.loc[next_d][fund_id_list] 
            fund_ret_n.name = 'ret'
            fund_pnl_i = pd.DataFrame([pos_n_1,nav_n_1,fund_ret_n]).T
            pnl_l = fund_pnl_i.prod(axis=1)
            pnl_l.name = d
            res.append(pnl_l)
        pnl_df = pd.DataFrame(res).fillna(0)
        desc_name_dict = fund_info.set_index('fund_id')[['desc_name']].to_dict()['desc_name']
        pnl_df.cumsum().rename(columns = desc_name_dict).plot.line(figsize=(18,9))
        plt.grid()
        plt.legend(fontsize=14 ,loc = 'lower left')
        plt.title('money loss during mdd period', fontsize=20)    
        plt.show()

    @staticmethod
    def plot_fund_ret_each_year(turnover_df:pd.DataFrame,
                                mdd: float,
                                annual_ret: float):
        annual_ret = round(annual_ret, 3)
        mdd = round(mdd, 3)
        ret_year = round(turnover_df.year_ret.mean(), 3)
        p = turnover_df.set_index('year')[['year_mdd','year_ret']].plot.bar(figsize=(12,9))
        t = plt.title('return and mdd each year', fontsize=20)
        s = plt.suptitle(f'annual ret: {annual_ret}, ret mean {ret_year},mdd {mdd}', y=0.87, fontsize=17)
        h = plt.axhline(y=annual_ret, linestyle=':', label='annual ret', color='darkorange')
        r = plt.xticks(rotation=0)
        l = plt.legend(fontsize=17 ,loc='upper left')
        plt.grid()
        
    @staticmethod
    def plot_turnover_rate_each_year(turnover_df:pd.DataFrame, turnover_rate_yearly_avg:float):
        turnover_rate_yearly_avg = round(turnover_rate_yearly_avg,2)
        p = turnover_df.set_index('year')[['turnover_rate_yearly']].plot.bar(figsize=(12,9))
        t = plt.title('turnover rate each year %', fontsize=20)
        s = plt.suptitle(f'mean {turnover_rate_yearly_avg}', y=0.87, fontsize=17)
        y = turnover_df['turnover_rate_yearly'].mean()
        h = plt.axhline(y=y, linestyle=':', label='turnover rate mean')
        r = plt.xticks(rotation=0)
        l = plt.legend(fontsize=17 ,loc='upper left')
        plt.grid()

    @staticmethod
    def plot_fund_score(fund_mv:pd.DataFrame, 
                        fund_weights_history:dict, 
                        trade_history:dict,
                        index_price:pd.DataFrame,
                        asset_weights:dict,
                        fund_info:pd.DataFrame,
                        fund_nav:pd.DataFrame,
                        fund_score:dict,
                        fund_score_raw:dict,
                        fund_indicator:pd.DataFrame,
                        asset=str,
                        is_tuning=bool,
                        ):
        fund_w = fund_weights_history
        end_date = fund_mv.index.tolist()[-1]
        date_list = list(trade_history.keys()) + [end_date]
        fund_asset_df = fund_info[['fund_id','index_id']].set_index('fund_id')
        fund_indicator = fund_indicator.pivot_table(index = ['fund_id','datetime'])
        traded_to_submit_date = {}
        for d in trade_history:
            dic = trade_history[d][0]
            traded_to_submit_date[dic.trade_date] = dic.submit_date
        res = []
        for k,v in asset_weights.items():
            v['date'] = k
            res.append(v)
        weights_df = pd.DataFrame(res).set_index('date')
        weights_df = weights_df.drop(['cash'], axis = 1).dropna()[1:]
        res = []
        for dic, s in zip(weights_df.to_dict('records'), weights_df.sum(axis = 1).values):
            res.append({ k: v/ s for k, v in dic.items()})
        df = pd.DataFrame(res)
        cols = df.columns.tolist()
        name_dic = fund_info[['fund_id','desc_name']].set_index('fund_id')
        b_d = fund_nav.index[0]
        e_d = fund_nav.index[-1]
        bench_df = index_price.loc[b_d:e_d,[asset]].dropna()
        bench_df = bench_df/bench_df.iloc[0]
        for i in range(len(date_list) - 1):   
            b_d = date_list[i]
            e_d = date_list[i+1]
            if b_d == e_d:
                break
            bench_df_tmp = bench_df.loc[b_d:e_d,:]
            bench_df_tmp = bench_df_tmp/bench_df_tmp.iloc[0]
            f_l = [k  for k ,v in fund_w[b_d].items() if isinstance(v, float) and round(v ,3 ) > 0]
            if 'cash' in f_l:
                f_l.remove('cash')
            f_l = [f for f in f_l if fund_asset_df.loc[f,'index_id'] == asset]
            mv_b = bench_df.loc[b_d,asset]
            fund_tmp = fund_nav.loc[b_d:e_d,f_l].copy()
            fund_tmp = fund_tmp/fund_tmp.iloc[0]
            fund_tmp = fund_tmp
            res = []
            submit_d = traded_to_submit_date[b_d]
            if len(f_l) < 1:
                continue
            fig, ax = plt.subplots(figsize= [16,12])
            plt.plot(bench_df_tmp.index, bench_df_tmp[asset], label=asset,linewidth=5.0)
            for f in f_l:
                desc = name_dic.loc[f,'desc_name']
                f_i_dict= fund_indicator.loc[f,submit_d].to_dict()
                dic = {
                    'fund_id' : f,
                    'desc_name' : desc,
                    'score' : round(fund_score[submit_d][asset][f],4),
                    'weight': round(fund_w[b_d][f], 4),
                    'score_raw': round(fund_score_raw[submit_d][asset][f],4),
                }
                for s in ['alpha','beta','fee_rate','track_err']:
                    dic[s] = round(f_i_dict[s],4)
                res.append(dic)    
                plt.plot(fund_tmp.index, fund_tmp[f], label=f+'_'+desc,linestyle='--',linewidth=3.0)    
            
            if is_tuning:
                fund_not_select = [f for f, v in fund_score[submit_d][asset].items() if f not in f_l][:10]
                if len(fund_not_select) < 1:
                    plt.legend(fontsize=17, loc = 'lower left')
                    plt.title(f'{asset} {b_d} {e_d}', fontsize=25)
                    plt.grid()
                    ax.xaxis.set_ticks_position('top')
                    continue
                fund_tmp = fund_nav.loc[b_d:e_d,fund_not_select].copy()
                fund_tmp = fund_tmp/fund_tmp.iloc[0]
                fund_tmp = fund_tmp
                for f in fund_not_select:
                    desc = name_dic.loc[f,'desc_name']
                    f_i_dict= fund_indicator.loc[f,submit_d].to_dict()
                    dic = {
                        'fund_id' : f,
                        'desc_name' : desc,
                        'score' : round(fund_score[submit_d][asset][f],4),
                        'weight': 0,
                        'score_raw': round(fund_score_raw[submit_d][asset][f],4),
                    }
                    for s in ['alpha','beta','fee_rate','track_err']:
                        dic[s] = round(f_i_dict[s],4)
                    res.append(dic)   
                    plt.plot(fund_tmp.index, fund_tmp[f], label=f+'_'+desc,linestyle=':',linewidth=3.0)    

            plt.legend(fontsize=17, loc = 'lower left')
            plt.title(f'{asset} {b_d} {e_d}', fontsize=25)
            #plt.suptitle(FundScoreManager().funcs[asset].__dict__, y=0.87, fontsize=18)
            ax.xaxis.set_ticks_position('top')
            fund_df = pd.DataFrame(res)
            fund_df = fund_df[['desc_name','fund_id','weight','alpha','beta','track_err','fee_rate','score_raw','score']]
            fund_df = fund_df.sort_values('score', ascending = False)
            t = plt.table(
                cellText=fund_df.values.tolist(),
                colLabels=fund_df.columns,
                loc='bottom',
                colWidths= [0.25,0.12,0.09,0.09,0.09,0.09,0.09,0.09,0.09]          
            )
            t.auto_set_font_size(False)
            t.set_fontsize(17)
            t.auto_set_column_width('fund_id')
            t.AXESPAD = 0.1
            t.scale(1, 4)
            plt.grid()
            plt.show()

    @staticmethod
    def plot_fund_alpha(fund_nav:pd.DataFrame, 
                        fund_info:pd.DataFrame, 
                        index_price:pd.DataFrame,
                        fund_position_history:dict,
                        backtest_result:dict):

        rebalance_date = backtest_result['rebalance_date'] + [backtest_result['market_value'].index.tolist()[-1]]
        fund_desc_name = fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']  
        res = {}
        for idx, d in enumerate(rebalance_date[:-1]):
            for fund_id in fund_position_history[d]:
                if fund_position_history[d][fund_id]['volume'] > 0:
                    index_id = fund_position_history[d][fund_id]['index_id']
                    dic_i = {
                        'start_date':d,
                        'fund_id': fund_id,
                        'volume':fund_position_history[d][fund_id]['volume'],
                        'end_date':rebalance_date[idx + 1],
                        'index_id':index_id,
                        'price':fund_position_history[d][fund_id]['price'],
                    }
                    if index_id in res:
                        res[index_id].append(dic_i)
                    else:
                        res[index_id] = [dic_i]
                    
        for index_id in res:
            df_index_funds = pd.DataFrame(res[index_id])
            df_index_funds['amount'] = df_index_funds['price'] * df_index_funds['volume']
            result = []
            red_box = []
            green_box = []
            for d_idx, d in enumerate(sorted(set(df_index_funds.start_date))):
                df_i  = df_index_funds[df_index_funds.start_date == d].copy()
                amount_sum = df_i.amount.sum()
                df_i.loc[:,'weight'] = df_i['amount'].map(lambda x: round(x / amount_sum,2))
                df_i = df_i.sort_values('weight').reset_index(drop=True)
                d1 = d
                d2 = df_i.end_date.values[0]
                index_ret = index_price.loc[d2,index_id] / index_price.loc[d1,index_id] - 1
                dic = {
                    'start_date':d1,
                    'end_date':d2,
                    index_id:int(10000*index_ret),
                }
                for _idx, r in df_i.iterrows():
                    f_r = int(10000 * (fund_nav.loc[d2,r.fund_id] / fund_nav.loc[d1,r.fund_id] - 1 - index_ret))
                    fund_loc = f'fund_{_idx+1}'
                    sign = "+" if f_r >= 0 else "-"
                    dic[fund_loc] = f'{fund_desc_name[r.fund_id]} {sign} {abs(f_r)}'
                    if f_r >= index_ret:
                        red_box.append((d_idx+1,_idx+1+2))
                    else:
                        green_box.append((d_idx+1,_idx+1+2))
                result.append(dic)
            df_result = pd.DataFrame(result)
            l = df_result.columns.tolist()
            l1 = ['start_date','end_date',index_id]
            l2 = [i for i in l if i not in l1]
            table_df = df_result[l1 + l2].fillna('')
            fig, ax = plt.subplots()
            ax.axis('tight')
            ax.axis('off')
            col_l = [0.2,0.2,0.15] + [0.60] * (table_df.shape[1]-3)
            the_table = ax.table(cellText=table_df.values.tolist(),
            colLabels=table_df.columns,
            colWidths= col_l,  
            loc='center')
            the_table.auto_set_font_size(False)
            the_table.set_fontsize(11)
            the_table.AXESPAD = 0.1
            the_table.scale(1, 3)
            for i in red_box:
                the_table[i].set_facecolor('#ffe5ad')
            for i in green_box:
                the_table[i].set_facecolor('#cffdbc')
            plt.show()

    @staticmethod
    def plot_mv_on_each_asset(fund_nav:pd.DataFrame, 
                                fund_info:pd.DataFrame, 
                                index_price:pd.DataFrame,
                                fund_position_history:dict,
                                backtest_result:dict):
        fund_info['trade_fee'] = fund_info['purchase_fee'] + fund_info['redeem_fee']
        fee_df = fund_info[['fund_id','trade_fee']].set_index('fund_id')

        rebalance_date = backtest_result['rebalance_date'] + [backtest_result['market_value'].index.tolist()[-1]]
        res = {}
        for idx, d in enumerate(rebalance_date[:-1]):
            for fund_id in fund_position_history[d]:
                if fund_position_history[d][fund_id]['volume'] > 0:
                    index_id = fund_position_history[d][fund_id]['index_id']
                    dic_i = {
                        'start_date':d,
                        'fund_id': fund_id,
                        'volume':fund_position_history[d][fund_id]['volume'],
                        'end_date':rebalance_date[idx + 1],
                        'index_id':index_id,
                        'price':fund_position_history[d][fund_id]['price'],
                    }
                    if index_id in res:
                        res[index_id].append(dic_i)
                    else:
                        res[index_id] = [dic_i]
                        
        for index_id in res:
            index_res = []
            df_index_funds = pd.DataFrame(res[index_id])
            df_index_funds['amount'] = df_index_funds['price'] * df_index_funds['volume']
            last_value = 1
            for d_idx, d in enumerate(sorted(set(df_index_funds.start_date))):
                df_i  = df_index_funds[df_index_funds.start_date == d].copy()
                amount_sum = df_i.amount.sum()
                df_i.loc[:,'weight'] = df_i['amount'].map(lambda x: x / amount_sum)
                d1 = d
                d2 = df_i.end_date.values[0]
                fund_sub_list = df_i.fund_id.tolist()
                index_df_i = fund_nav.loc[d1:d2][df_i.fund_id.tolist()]
                for fund_id in fund_sub_list:
                    fund_i_fee = fee_df.loc[fund_id,'trade_fee']
                    fee_discount_l = np.linspace(1,1-fund_i_fee,index_df_i.shape[0])
                    index_df_i[fund_id] = index_df_i[fund_id] * fee_discount_l
                w_df = df_i[['fund_id','volume']].set_index('fund_id').T
                weighted_s = (index_df_i * w_df.values).sum(axis=1)
                ws = weighted_s.iloc[:-1]
                ws = ws / ws[0] * last_value
                last_value = ws[-1] 
                index_res.append(ws)
            
            asset_weighted_index = pd.DataFrame(pd.concat(index_res)).rename(columns={0:index_id+' fund weighted sum'})
            asset_weighted_index = index_price[[index_id]].join(asset_weighted_index).fillna(method='ffill').dropna()
            asset_weighted_index = asset_weighted_index / asset_weighted_index.iloc[0]
            asset_weighted_index.plot.line(figsize=(15,12))
            last_value = round(asset_weighted_index[index_id+' fund weighted sum'].values[-1],4)
            plt.title(f'{index_id} funds weighted price vs {index_id} ', fontsize=25)
            plt.suptitle(f'last asset value {last_value}', y=0.87, fontsize=17)
            plt.legend(fontsize=17, loc = 'lower left')
            plt.grid()
            plt.show()

    @staticmethod
    def plot_recent_ret(df_plot_i:pd.DataFrame,begin_date:datetime.date,end_date:datetime.date,st:str,port_id:str):
        df_plot_i.index = [str(i) for i in df_plot_i.index]
        df_plot_i.plot.line(figsize=(12,6))
        plt.title(f'portfolio {port_id} mv from {begin_date} to {end_date}', fontsize=20)
        plt.suptitle(st,y=0.87,fontsize=17)
        plt.legend(loc='lower left',fontsize = 15)
        plt.grid()
        plt.show()

    @staticmethod
    def plot_fund_rank( score_dic:pd.DataFrame, 
                        fund_info:pd.DataFrame, 
                        fund_id:str,
                        begin_date:datetime.date,
                        fund_index_map:dict):
        index_id = fund_index_map[fund_id]
        res = []
        for d in score_dic:
            if d > begin_date:
                score_d = score_dic[d][index_id]
                sorted_score = sorted(score_d.items(), key=lambda x:x[1], reverse=True)
                sorted_score = [i[0] for i in sorted_score]
                try:
                    rank = sorted_score.index(fund_id)
                except:
                    rank = 666
                res.append({'datetime':d,'rank':rank})
        desc_name = fund_info[fund_info.fund_id == fund_id].desc_name.values[0]
        score = pd.DataFrame(res).set_index('datetime')
        score.plot.line(figsize=(12,6))
        plt.title(f'{fund_id}  {desc_name}  rank  ', fontsize=25)
        plt.grid()
        plt.show()

    @staticmethod
    def plot_rolling_analysis( annual_ret:pd.DataFrame, 
                               fund_annual_ret:float,
                               year:int,
                               index_id:'str'):
        
        df1 = annual_ret.dropna()[[index_id,'战略配置']]
        index = df1.index
        saa1 = df1[index_id]
        taa1 = df1.战略配置
        df2 = annual_ret.dropna()[['战略配置','战术配置']]
        index = df2.index
        saa2 = df2.战略配置
        taa2 = df2.战术配置
        df3 = annual_ret.dropna()[['战术配置','优选基金']]
        index = df3.index
        saa3 = df3.战术配置
        taa3 = df3.优选基金
        df4 = annual_ret.dropna()[[index_id,'优选基金']]
        index = df4.index
        saa4 = df4[index_id]
        taa4 = df4.优选基金

        df1.plot.line(figsize=(16,6))
        plt.fill_between(index, saa1, taa1,
                        where=(saa1 < taa1),
                        alpha=0.30, color='#ff7f0e', interpolate=False)
        plt.fill_between(index, saa1, taa1,
                        where=(saa1 > taa1),
                        alpha=0.30, color='#1f77b4', interpolate=False)
        pl.legend(loc='upper left',fontsize = 17)
        pl.title(f'滚动{year}年化收益对比', fontsize=20)
        plt.grid()
        plt.show()
        
        df2.plot.line(figsize=(16,6))
        plt.fill_between(index, saa2, taa2,
                        where=(saa2 < taa2),
                        alpha=0.30, color='#ff7f0e', interpolate=False)
        plt.fill_between(index, saa2, taa2,
                        where=(saa2 > taa2),
                        alpha=0.30, color='#1f77b4', interpolate=False)
        pl.legend(loc='upper left',fontsize = 17)
        pl.title(f'滚动{year}年化收益对比', fontsize=20)
        plt.grid()
        plt.show()
        
        df3.plot.line(figsize=(16,6))
        plt.fill_between(index, saa3, taa3,
                        where=(saa3 < taa3),
                        alpha=0.30, color='#ff7f0e', interpolate=False)
        plt.fill_between(index, saa3, taa3,
                        where=(saa3 > taa3),
                        alpha=0.30, color='#1f77b4', interpolate=False)
        pl.legend(loc='upper left',fontsize = 17)
        pl.title(f'滚动{year}年化收益对比', fontsize=20)
        plt.grid()
        plt.show()
        df4.plot.line(figsize=(16,6))
        plt.fill_between(index, saa4, taa4,
                        where=(saa4 < taa4),
                        alpha=0.30, color='#ff7f0e', interpolate=False)
        plt.fill_between(index, saa4, taa4,
                        where=(saa4 > taa4),
                        alpha=0.30, color='#1f77b4', interpolate=False)
        pl.legend(loc='upper left',fontsize = 17)
        pl.title(f'滚动{year}年化收益对比', fontsize=20)
        plt.grid()
        plt.show()
        df_funds = annual_ret.dropna()[['优选基金']].rename(columns={'优选基金':'value'})
        df_funds.plot.hist(bins=20, figsize=(16,10),legend=None)
        plt.title(f'基金滚动{year}年收益分布', fontsize=25)
        plt.grid()
        plt.show()
        df_funds = df_funds.groupby('value')['value'].agg('count').pipe(pd.DataFrame).rename(columns = {'value': 'frequency'})
        df_funds['pdf'] = df_funds['frequency'] / sum(df_funds['frequency'])
        df_funds['cdf'] = df_funds['pdf'].cumsum()
        df_funds.index.name = '年化收益'
        index_loc = df_funds.index.values[df_funds.index.values > fund_annual_ret][0]
        cdf_loc = df_funds.loc[index_loc, 'cdf']
        df_funds = 1 - df_funds[['cdf']]
        df_funds[['cdf']].plot.line(figsize=(12,6), legend=None)
        mean_ret = round(100 * abs(df_funds['cdf'] - 0.5).idxmin(),2)
        plt.scatter(index_loc, 1-cdf_loc, marker='D', color='C1', s=60)
        plt.title(f'{year}年年化收益 累积密度分布', fontsize=20)
        rate = round((1-cdf_loc)*100,2)
        st = f'组合年化收益 {round(fund_annual_ret * 100, 2)}%  获取年化收益胜率 {rate}%  50%胜率年化收益 {mean_ret}%'
        plt.suptitle(st,y=0.87,fontsize=17)
        plt.grid()
        plt.show()
        


import pandas as pd
import dataclasses
import datetime


@dataclasses.dataclass
class FundDataTables:
    # raw
    cs_index_component: pd.DataFrame = None
    # basic
    trading_days: pd.DataFrame = None
    index_info: pd.DataFrame = None
    index_price: pd.DataFrame = None
    index_pct: pd.DataFrame = None
    fund_info: pd.DataFrame = None
    fund_score: pd.DataFrame = None
    fund_nav: pd.DataFrame = None
    fund_indicator: pd.DataFrame = None
    fund_unit_nav: pd.DataFrame = None
    fund_size: pd.DataFrame = None
    fund_com_hold: pd.DataFrame = None
    # fund helps
    fund_list: set = None
    index_list: set = None
    fund_index_map: dict = None # fund_id -> index_id
    # score helps
    index_fund_indicator_pack: pd.DataFrame = None # pivot_table

    def __post_init__(self):
        self.fund_list = self.fund_list or set([])
        self.fund_index_map = self.fund_index_map or {}

    def __repr__(self):
        s = '<DTS'
        for k, v in self.__dict__.items():
            s += f' {k}:{v.shape if isinstance(v, pd.DataFrame) else (len(v) if isinstance(v, set) or isinstance(v, list) else None)}'
        s += '>'
        return s

    def _agg_indicator(self):
        self.fund_indicator = self.fund_indicator.rename(columns={'var':'var_'})
        self.index_fund_indicator_pack = self.fund_indicator.pivot_table(index=['index_id', 'datetime', 'fund_id'])
        self.index_list = set(self.index_fund_indicator_pack.index.levels[0])

    def fix_data(self):
        # 将 fix_data_v1 去掉的原因是目前 basic 改用choice数据源，nav数据没有这个问题了。
        # self.fix_data_v1()
        self._agg_indicator()

    def fix_data_v1(self):
        '''
        这个改动主要原因是米筐数据源在nav的处理中有问题，导致一些被选中的基金受到影响，所以采用这个fix
        '''
        # ----- after load data_table, we need to do some process
        def change_nav_element(fund_id, d, rate):
            if fund_id in self.fund_nav.columns:
                self.fund_nav.loc[d:,fund_id] = self.fund_nav.loc[d:,fund_id] * rate

        #人工改数据
        fund_id = '003520!0'
        d = datetime.date(2019,3,24)
        rate = 1.0883/1.0013/1.0879*1.089826
        change_nav_element(fund_id, d, rate)

        # 泰达宏利沪深300指数增强A
        # rate number:  1.4287351257399752
        fund_id = '162213!0'
        d = datetime.date(2018,3,16)
        rate = 1.9315/1.4915/1.943*2.143646
        change_nav_element(fund_id, d, rate)

        # 万家1-3年政策性金融债C
        # rate number:  1.0931219248656086
        fund_id = '003521!0'
        d = datetime.date(2019,3,25)
        rate = 1.0916/1.0006/1.0912*1.093377
        change_nav_element(fund_id, d, rate) 